#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import itertools
import json
import os
import time

from naruno.config import UNL_NODES_PATH
from naruno.lib.config_system import get_config
from naruno.lib.kot import KOT

unl_db = KOT("unl", folder=get_config()["main_folder"] + "/db")


class Unl:

    @staticmethod
    def save_new_unl_node(node_id):
        """
        Saves the new unl.
        """

        nodes_list = Unl.get_unl_nodes()

        already_in_list = any(element == node_id for element in nodes_list)
        if not already_in_list:
            nodes_list[node_id] = {}
            nodes_list[node_id]["date"] = time.time()

            unl_db.set("unl", nodes_list)

    @staticmethod
    def get_unl_nodes(custom_UNL_NODES_PATH=None):
        """
        Returns the UNL nodes list from UNL_NODES_PATH.
        """

        record = (unl_db.get("unl") if custom_UNL_NODES_PATH is None else KOT(
            "unl" + custom_UNL_NODES_PATH,
            folder=get_config()["main_folder"] + "/db",
        ).get("unl"))
        return record if record is not None else {}

    @staticmethod
    def get_as_node_type(id_list, c_type=0):
        """
        Converts the UNL node list to Node class.
        """

        from naruno.node.server.server import server
        nodes = []
        if server.Server is not None:
         for i in server.Server.clients:
            if i.c_type == c_type:
                nodes.append(i)
        return nodes

    @staticmethod
    def node_is_unl(node_id):
        """
        Returns the this node is unl or not.
        """

        for unl in Unl.get_unl_nodes():
            temp_unl = unl
            if node_id == temp_unl:
                return True
        return False

    @staticmethod
    def unl_node_delete(node_id):
        """
        Deletes the UNL node
        """

        saved_nodes = Unl.get_unl_nodes()
        if node_id in saved_nodes:
            del saved_nodes[node_id]

            unl_db.set("unl", saved_nodes)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import os

from naruno.config import COMMANDERS_PATH
from naruno.lib.config_system import get_config
from naruno.lib.kot import KOT

commanders_db = KOT("commanders", folder=get_config()["main_folder"] + "/db")


def DeleteCommander(tx):
    if commanders_db.get(tx) == True:
        commanders_db.delete(tx)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import time
import copy
from naruno.blockchain.block.block_main import Block
from naruno.blockchain.candidate_block.candidate_block_main import \
    candidate_block
from naruno.node.unl import Unl
import naruno

our_candidates = []

def self_candidates(block: Block, reset=False):
            the_block = copy.copy(block)
            the_block_2 = copy.copy(block)
            if not len(naruno.node.get_candidate_blocks.our_candidates) == 0 and not reset:
                the_block = naruno.node.get_candidate_blocks.our_candidates[2]
                the_block_2 = naruno.node.get_candidate_blocks.our_candidates[3]

            new_list = []
            signature_list = []
            a_time = "self"
            for element in block.validating_list:
                new_list.append(element.dump_json())
                signature_list.append(element.signature)
            will_add_candidate_block = None
            will_add_candidate_block_hash = None
            if not len(naruno.node.get_candidate_blocks.our_candidates) == 0 and not reset:
                will_add_candidate_block = naruno.node.get_candidate_blocks.our_candidates[0]
                will_add_candidate_block_hash = naruno.node.get_candidate_blocks.our_candidates[1]
            
            first_validating = [i.dump_json() for i in the_block.validating_list]
            second_validating = [i.dump_json() for i in block.validating_list]
            if not block.round_1 or will_add_candidate_block is None:
                if block.sequence_number+block.empty_block_number > the_block.sequence_number+the_block.empty_block_number or block.sequence_number == 0 or reset or will_add_candidate_block is None:

                    the_block = copy.copy(block)
                    will_add_candidate_block = {
                                "action": "myblock",
                                "transaction": new_list,
                                "signature": a_time,
                                "sequence_number": the_block.sequence_number+the_block.empty_block_number,
                                "total_length": len(new_list)
                            }

            if (block.round_1 and not block.round_2)  or block.sequence_number == 0 or will_add_candidate_block_hash is None:
                if block.sequence_number+block.empty_block_number > the_block_2.sequence_number+the_block_2.empty_block_number or block.sequence_number == 0 or reset or will_add_candidate_block_hash is None:
                    last_control = True
                    if block.sequence_number != 0:
                        if block.hash == None:
                            last_control = False
            

                    if last_control:

                        the_block_2 = copy.copy(block)
                    
                        will_add_candidate_block_hash = {
                                    "action":
                                    "myblockhash",
                                    "hash":
                                    the_block_2.hash,
                                    "previous_hash":
                                    the_block_2.previous_hash,
                                    "signature":
                                    a_time,
                                    "sequence_number":
                                    the_block_2.sequence_number+the_block_2.empty_block_number,
                                }
                


            
            naruno.node.get_candidate_blocks.our_candidates = [will_add_candidate_block, will_add_candidate_block_hash, the_block, the_block_2]
            

            return [the_block, the_block_2]


def our_candidates_f(block: Block, reset=False):
    
    self_candidates(block, reset=reset)
    return naruno.node.get_candidate_blocks.our_candidates


def GetCandidateBlocks(custom_nodes_list=None, block: Block = None, reset=False):
    """
    Collects candidate blocks and candidate block hashes
    from connected unl nodes and returns them in the
    candidate_block class
    """

    nodes = (Unl.get_as_node_type(Unl.get_unl_nodes()) + Unl.get_as_node_type(Unl.get_unl_nodes(),c_type=3)
             if custom_nodes_list is None else custom_nodes_list)

    the_candidate_blocks = []
    the_candidate_block_hashes = []
    id_control_list = []

    for node in nodes:
        if node.candidate_block is not None:
            the_id = ""
            if int(node.candidate_block["sequence_number"]
                   ) == block.sequence_number+block.empty_block_number:
                the_id = node.candidate_block["id"]
                if not the_id in id_control_list:
                    the_candidate_blocks.append(node.candidate_block)
            else:
                for i in node.candidate_block_history:
                    if i["sequence_number"] == block.sequence_number+block.empty_block_number:
                        the_id = i["id"]
                        if not the_id in id_control_list:
                            the_candidate_blocks.append(i)
                        
            if not the_id in id_control_list:
                id_control_list.append(the_id)
        else:
            pass
        if node.candidate_block_hash is not None:
            if (int(node.candidate_block_hash["sequence_number"]) ==
                    block.sequence_number+block.empty_block_number):
                the_candidate_block_hashes.append(node.candidate_block_hash)
            else:
                for i in node.candidate_block_hash_history:
                    if i["sequence_number"] == block.sequence_number+block.empty_block_number:
                        the_candidate_block_hashes.append(i)
        else:
            pass

    if block is not None:
        the_candidates = our_candidates_f(block, reset=reset)

        the_candidate_blocks.append(the_candidates[0])
        the_candidate_block_hashes.append(the_candidates[1])


    not_none_the_candidate_blocks = []

    for none_candidate_block in the_candidate_block_hashes:
        if not none_candidate_block["hash"] == None:
            not_none_the_candidate_blocks.append(none_candidate_block)

    return candidate_block(the_candidate_blocks, not_none_the_candidate_blocks)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import contextlib
import json
import os

from naruno.blockchain.block.block_main import Block
from naruno.config import TEMP_BLOCK_PATH
from naruno.consensus.rounds.round_1.process.transactions.checks.duplicated import \
    Remove_Duplicates
from naruno.lib.config_system import get_config
from naruno.lib.log import get_logger
import naruno
logger = get_logger("BLOCKCHAIN")

the_ram_block = {}

def GetBlock(custom_TEMP_BLOCK_PATH=None,
             get_normal_block=False,
             dont_clean=False, reset=False):
    """
    Returns the block.
    """
    the_TEMP_BLOCK_PATH = (TEMP_BLOCK_PATH if custom_TEMP_BLOCK_PATH is None
                           else custom_TEMP_BLOCK_PATH)

    if naruno.blockchain.block.get_block.the_ram_block is not None and not reset:
        if the_TEMP_BLOCK_PATH in naruno.blockchain.block.get_block.the_ram_block:
            return naruno.blockchain.block.get_block.the_ram_block[the_TEMP_BLOCK_PATH]

    from naruno.blockchain.block.save_block import block_db

    logger.debug("Getting block from disk")



    os.chdir(get_config()["main_folder"])

    highest_the_TEMP_BLOCK_PATH = the_TEMP_BLOCK_PATH
    highest_number = 0

    highest_second_number = 0
    highest_other_high_number = 0
    for file in os.listdir("db/"):
        if ("db/" + file).startswith(the_TEMP_BLOCK_PATH) and not (
                "db/" + file) == the_TEMP_BLOCK_PATH:
            with contextlib.suppress(IndexError):
                number = int((("db/" + file).replace(the_TEMP_BLOCK_PATH,
                                                    "")).split("-")[1])  # seq
                high_number = int(
                    (("db/" + file).replace(the_TEMP_BLOCK_PATH,
                                            "")).split("-")[3])  # val
                other_high_number = int(
                    (("db/" + file).replace(the_TEMP_BLOCK_PATH,
                                            "")).split("-")[2])  # val

                # Write a code for getting the blocks with high number
                if number > highest_number:
                    highest_number = number
                    highest_the_TEMP_BLOCK_PATH = "db/" + file
                    highest_second_number = high_number
                    highest_other_high_number = other_high_number
                elif number == highest_number:
                    if high_number > highest_second_number:
                        highest_number = number
                        highest_the_TEMP_BLOCK_PATH = "db/" + file
                        highest_second_number = high_number
                        highest_other_high_number = other_high_number
                    elif high_number == highest_second_number:
                        if other_high_number > highest_other_high_number:
                            highest_number = number
                            highest_the_TEMP_BLOCK_PATH = "db/" + file
                            highest_second_number = high_number
                            highest_other_high_number = other_high_number

    logger.debug("Highest block: " + highest_the_TEMP_BLOCK_PATH)

    result_normal = Block("non")
    block_db_path_first = os.path.join(get_config()["main_folder"],
                                       the_TEMP_BLOCK_PATH)
    record_of_normal = Block.load_json(block_db.get(the_TEMP_BLOCK_PATH,
                                    custom_key_location=block_db_path_first))
    if record_of_normal is not None:
        result_normal = record_of_normal

    block_db_path_second = os.path.join(get_config()["main_folder"],
                                        highest_the_TEMP_BLOCK_PATH)
    logger.debug("Highest block path: " + block_db_path_second)
    result_highest = Block.load_json(block_db.get(highest_the_TEMP_BLOCK_PATH,
                                  custom_key_location=block_db_path_second))


    result_normal = Remove_Duplicates(result_normal)
    result_highest = Remove_Duplicates(result_highest)

    result_normal.validating_list = sorted(result_normal.validating_list,
                                           key=lambda x: x.fromUser)

    result_highest.validating_list = sorted(result_highest.validating_list,
                                            key=lambda x: x.fromUser)

    if get_normal_block:
        return result_normal

    if (result_normal.sequence_number + result_normal.empty_block_number
            > result_highest.sequence_number +
            result_highest.empty_block_number):
        return result_normal
    elif (result_normal.sequence_number +
          result_normal.empty_block_number == result_highest.sequence_number +
          result_highest.empty_block_number):
        result_normal_situation = 0
        result_highest_situation = 0
        if result_normal.round_1:
            result_normal_situation += 1
        if result_normal.round_2:
            result_normal_situation += 1

        if result_highest.round_1:
            result_highest_situation += 1
        if result_highest.round_2:
            result_highest_situation += 1

        if len(result_normal.validating_list) > len(
                result_highest.validating_list):
            return result_normal
        else:
            return result_highest

    else:
        return result_highest



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import json
import os

from naruno.config import WALLETS_PATH
from naruno.lib.config_system import get_config
from naruno.lib.kot import KOT
wallet_db = KOT("wallet",
                        folder=get_config()["main_folder"] + "/db")

def get_saved_wallet():
    record = wallet_db.get("wallet")


    return record if record is not None else {}



#!/usr/bin/python3
# -*- coding: utf-8 -*-
import os
import shutil
import time

from kivy.app import App
from kivy.core.clipboard import Clipboard
from kivy.utils import platform
from kivymd.uix.button import MDFlatButton
from kivymd.uix.button import MDRaisedButton
from kivymd.uix.gridlayout import MDGridLayout
from kivymd.uix.screen import MDScreen
from kivymd_extensions.sweetalert import SweetAlert

import naruno.gui.the_naruno_gui_app
from naruno.gui.popup import popup
from naruno.lib.backup.naruno_export import naruno_export
from naruno.lib.backup.naruno_import import naruno_import
from naruno.lib.settings_system import d_mode_settings
from naruno.lib.settings_system import dark_mode_settings
from naruno.lib.settings_system import mt_settings
from naruno.lib.settings_system import t_mode_settings
from naruno.lib.settings_system import the_settings
from naruno.lib.settings_system import baklava_settings


class SettingsScreen(MDScreen):
    pass


class SettingsBox(MDGridLayout):
    cols = 2

    d_first_status = the_settings()["debug_mode"]
    t_first_status = the_settings()["test_mode"]
    mt_first_status = the_settings()["mute_notifications"]
    dark_mode_first_status = the_settings()["dark_mode"]
    baklava_first_status = the_settings()["baklava"]

    def D_Status_Changing(self, instance, value):
        d_mode_settings(value)

    def T_Status_Changing(self, instance, value):
        t_mode_settings(value)

    def MT_Status_Changing(self, instance, value):
        mt_settings(value)

    def show_dialog(self):
        button_ok = MDRaisedButton(
            text="OK",
            font_size=16,
            on_release=self.callback,
        )
        button_cancel = MDFlatButton(
            text="CANCEL",
            font_size=16,
            on_release=self.callback,
        )
        self.alert = SweetAlert()
        self.alert.fire(
            "The app will restart.",
            buttons=[button_ok, button_cancel],
            type="info",
        )

    def callback(self, instance_button):
        if instance_button.text == "OK":
            self.alert.dismiss()
            SettingsBox.dark_mode_first_status = the_settings()["dark_mode"]
            naruno.gui.the_naruno_gui_app.the_naruno_gui.restart()

        else:
            self.alert.dismiss()

    def DARK_MODE_Status_Changing(self, instance, value):
        dark_mode_settings(value)
        self.show_dialog()

    def BAKLAVA_Status_Changing(self, instance, value):
        baklava_settings(value)


    def export_bt(self):
        export_location = naruno_export()
        Clipboard.copy(export_location)
        if platform == "android":
            from android.storage import primary_external_storage_path

            dir = primary_external_storage_path()
            download_dir_path = os.path.join(dir, "Download")
            new_path = os.path.join(download_dir_path,
                                    export_location.split("/")[-1])
            Clipboard.copy(new_path)
            shutil.move(
                export_location,
                new_path,
            )
        popup(
            title="The export file location has been copied to your clipboard.",
            type="success",
        )

    def import_the_db(self):
        naruno_import(self.import_backup_dialog.input_results["Path"])
        SettingsBox.dark_mode_first_status = the_settings()["dark_mode"]
        SettingsBox.d_first_status = the_settings()["debug_mode"]
        SettingsBox.t_first_status = the_settings()["test_mode"]
        SettingsBox.mt_first_status = the_settings()["mute_notifications"]
        SettingsBox.baklava_first_status = the_settings()["baklava"]
        naruno.gui.the_naruno_gui_app.the_naruno_gui.restart()

    def import_bt(self):
        self.import_backup_dialog = popup(
            title="Import your export (App Will Restart)",
            target=self.import_the_db,
            inputs=[["Path", False]],
        )



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import contextlib
import copy
import json
import os
import traceback
from urllib.request import urlopen

from naruno.lib.config_system import get_config
from naruno.lib.kot import KOT
from naruno.lib.settings_system import the_settings
from naruno.transactions.transaction import Transaction
import naruno

mytransactions_db = KOT("mytransactions",
                        folder=get_config()["main_folder"] + "/db")

mytransactions_db_ram = {}


def check_from_network():
    """
    Checks if the transaction is in the network.
    """
    validated_transactions = []
    if the_settings()["baklava"]:
        # export validated transactions
        response = (
            urlopen("http://test_net.1.naruno.org:8000/transactions/received"
                    ).read().decode("utf-8"))
        response = json.loads(response)
        for transaction in response:
            if response[transaction]["validated"]:
                validated_transactions.append(transaction)

    return validated_transactions


def GetMyTransaction(sended=None, validated=None, turn_json=False) -> list:
    """
    Returns the transaction db.
    """
    network_validated_source = check_from_network()

    network_validated = []

    the_transactions = []

    if len(naruno.transactions.my_transactions.get_my_transaction.mytransactions_db_ram) != mytransactions_db.get_count():
        naruno.transactions.my_transactions.get_my_transaction.mytransactions_db_ram = {}

    all_records = (mytransactions_db.get_all()
                   if naruno.transactions.my_transactions.get_my_transaction.mytransactions_db_ram == {} else naruno.transactions.my_transactions.get_my_transaction.mytransactions_db_ram)
    for entry in copy.copy(all_records):
        if not entry.endswith("validated") and not entry.endswith("sended"):
            try:
                the_transactions_json = all_records[entry]
                the_tx = Transaction.load_json(the_transactions_json)
                each_validated = (False if mytransactions_db.get(entry +
                                                                 "validated")
                                  == None else True)

                if (the_transactions_json["signature"]
                        in network_validated_source and not each_validated):
                    each_validated = True
                    network_validated.append(the_tx)

                each_sended = (False if mytransactions_db.get(entry + "sended")
                               == None else True)
                the_transactions.append([
                    the_tx,
                    each_validated,
                    each_sended,
                ])
            except json.decoder.JSONDecodeError:
                mytransactions_db.delete(entry)
                mytransactions_db.delete(entry + "validated")
                mytransactions_db.delete(entry + "sended")

    if sended is not None:
        the_transactions = [tx for tx in the_transactions if tx[2] == sended]

    if validated is not None:
        the_transactions = [
            tx for tx in the_transactions if tx[1] == validated
        ]

    # sort
    the_transactions.sort(key=lambda x: x[0].signature)

    from naruno.transactions.my_transactions.validate_transaction import \
        ValidateTransaction

    for i in network_validated:
        ValidateTransaction(i,
                            custom_currently_list=the_transactions,
                            force_notify=True)

    if turn_json:
        the_transactions = {
            tx[0].signature: {
                "transaction": tx[0].dump_json(),
                "validated": tx[1],
                "sended": tx[2],
            }
            for tx in the_transactions
        }

    return the_transactions



#!/usr/bin/python3
# -*- coding: utf-8 -*-
import base64
import contextlib
import hashlib
import os
import pickle
import time
import traceback
from datetime import datetime
from hashlib import sha256
from shutil import copy
from shutil import make_archive
from shutil import move
from shutil import rmtree
from shutil import unpack_archive
import random


force_compress = False
force_encrypt = False


class HASHES:
    cache = {}
    @staticmethod
    def get_hash(string: str, hash_type: str = "sha256") -> str:
        if string in HASHES.cache:
            return HASHES.cache[string]
        if hash_type == "sha256":
            result = sha256(string.encode()).hexdigest()
        elif hash_type == "sha512":
            result = sha512(string.encode()).hexdigest()
        elif hash_type == "md5":
            result = md5(string.encode()).hexdigest()
        else:
            raise ValueError("Hash type must be sha256, sha512 or md5")
        HASHES.cache[string] = result
        return result


class KOT:

    @staticmethod
    def benchmark_set(number: int = 10000,
                      compress: bool = False,
                      encryption_key: str = "") -> float:
        compress = True if force_compress else compress
        encryption_key = force_encrypt if force_encrypt != False else encryption_key
            
        my_db = KOT("KOT-benchmark", self_datas=True)
        start = time.time()
        for i in range(number):
            my_db.set(
                "key" + str(i),
                "value" + str(i),
                compress=compress,
                encryption_key=encryption_key,
            )
        end = time.time()
        my_db.delete_all()
        return end - start

    @staticmethod
    def benchmark_get(
        number: int = 10000,
        compress: bool = False,
        encryption_key: str = "",
        dont_generate: bool = False,
    ) -> float:
        compress = True if force_compress else compress
        encryption_key = force_encrypt if force_encrypt != False else encryption_key
        my_db = KOT("KOT-benchmark", self_datas=True)
        if not dont_generate:
            for i in range(number):
                my_db.set(
                    "key" + str(i),
                    "value" + str(i),
                    compress=compress,
                    encryption_key=encryption_key,
                )
        start = time.time()
        for i in range(number):
            my_db.get("key" + str(i), encryption_key=encryption_key)
        end = time.time()
        my_db.delete_all()
        return end - start

    @staticmethod
    def benchmark_delete(
        number: int = 10000,
        compress: bool = False,
        encryption_key: str = "",
        dont_generate: bool = False,
    ) -> float:
        compress = True if force_compress else compress
        encryption_key = force_encrypt if force_encrypt != False else encryption_key
        my_db = KOT("KOT-benchmark", self_datas=True)
        if not dont_generate:
            for i in range(number):
                my_db.set(
                    "key" + str(i),
                    "value" + str(i),
                    compress=compress,
                    encryption_key=encryption_key,
                )
        start = time.time()
        for i in range(number):
            my_db.delete("key" + str(i))
        end = time.time()
        my_db.delete_all()
        return end - start

    @staticmethod
    def benchmark(number: int = 10000,
                  compress: bool = False,
                  encryption_key: str = "") -> float:
        compress = True if force_compress else compress
        encryption_key = force_encrypt if force_encrypt != False else encryption_key
        total_time = 0
        total_time += KOT.benchmark_set(number, compress, encryption_key)
        total_time += KOT.benchmark_get(number,
                                        compress,
                                        encryption_key,
                                        dont_generate=True)
        total_time += KOT.benchmark_delete(number,
                                           compress,
                                           encryption_key,
                                           dont_generate=True)
        return total_time

    @staticmethod
    def database_list(folder: str = "") -> dict:
        database_index = KOT("KOT-database-index",
                             self_datas=True,
                             folder=folder)
        return database_index.dict()
    @staticmethod
    def gui(password, folder: str = ""):
        from .gui import GUI
        GUI(folder, password)


    @staticmethod
    def web(password, folder: str = "",host=None, port=0):
        from .gui import WEB
        WEB(folder, password, host, port)

    @staticmethod
    def database_delete(name: str, folder: str = "") -> bool:
        database_index = KOT("KOT-database-index",
                             self_datas=True,
                             folder=folder)
        try:
            the_db = KOT(name, folder=folder, self_datas=True)
            for each_key in the_db.get_all():
                the_db.delete(each_key)
            rmtree(database_index.get(name))
        except:
            return False

        database_index.delete(name)
        return True


    @staticmethod
    def database_pop(name: str, folder: str = "") -> bool:
        database_index = KOT("KOT-database-index",
                             self_datas=True,
                             folder=folder)
        try:
            the_db = KOT(name, folder=folder, self_datas=True)
            for each_key in the_db.get_all():
                the_db.delete(each_key)
        except:
            return False

        return True

    @staticmethod
    def database_delete_all(folder: str = ""):
        database_index = KOT("KOT-database-index",
                             self_datas=True,
                             folder=folder)

        for each_database in database_index.dict():
            KOT.database_delete(each_database, folder=folder)

    @staticmethod
    def database_pop_all(folder: str = ""):
        database_index = KOT("KOT-database-index",
                             self_datas=True,
                             folder=folder)

        for each_database in database_index.dict():
            KOT.database_pop(each_database, folder=folder)

    @staticmethod
    def database_rename(name: str,
                        new_name: str,
                        force: bool = False,
                        folder: str = "") -> bool:
        if new_name in KOT.database_list() and not force:
            return False
        try:
            first_db = KOT(name, folder=folder)
            location = first_db.backup(".")
            KOT.database_delete(name)
            second_db = KOT(new_name, folder=folder)
            second_db.restore(location)
            os.remove(location)
        except:
            traceback.print_exc()
            return False
        return True

    def __init__(self, name, self_datas: bool = False, folder: str = ""):
        self.name = name
        self.hashed_name = HASHES.get_hash(name)
        the_main_folder = os.getcwd() if not folder != "" else folder
        self.the_main_folder = the_main_folder
        self.location = os.path.join(the_main_folder,
                                     "KOT-" + self.hashed_name)

        if not self_datas:
            self.open_files_db = KOT("KOT-open_files_db",
                                     self_datas=True,
                                     folder=folder)
            database_index = KOT("KOT-database-index",
                                 self_datas=True,
                                 folder=folder)
            database_index.set(self.name, self.location)

        self.counter = 0

        self.cache = {}

        self.initialize()

    def initialize(self):
        try:
            os.makedirs(self.location)
        except OSError:
            if not os.path.isdir(self.location):
                raise

    def clear_cache(self):
        self.cache = {}
        for each_file in self.open_files_db.dict():
            if os.path.exists(each_file):
                os.remove(each_file)
        self.open_files_db.delete_all()

    def encrypt(self, key, message):
        from Crypto.Cipher import AES

        # Serializing the message
        message = pickle.dumps(message)
        # Turning to string
        message = base64.b64encode(message).decode("ascii")

        def pad(s):
            return s + (16 - len(s) % 16) * chr(16 - len(s) % 16)

        padded_message = pad(message)
        from Crypto import Random

        iv = Random.new().read(AES.block_size)
        cipher = AES.new(
            hashlib.sha256(key.encode()).digest(), AES.MODE_CBC, iv)
        return base64.b64encode(
            iv + cipher.encrypt(padded_message.encode())).decode()

    def decrypt(self, key, message):
        from Crypto.Cipher import AES

        def unpad(s):
            return s[:-ord(s[len(s) - 1:])]

        message = base64.b64decode(message.encode())
        iv = message[:AES.block_size]
        cipher = AES.new(
            hashlib.sha256(key.encode()).digest(), AES.MODE_CBC, iv)
        unpadded = unpad(cipher.decrypt(message[AES.block_size:])).decode()

        unpadded = base64.b64decode(unpadded)

        return pickle.loads(unpadded)

    def set(
        self,
        key: str,
        value=None,
        file: str = "",
        compress: bool = False,
        encryption_key: str = "",
        cache_policy: int = 0,
        dont_delete_cache: bool = False,
        dont_remove_file: bool = False,
        custom_key_location: str = "",
        short_cut: bool = False,
    ) -> bool:
        compress = True if force_compress else compress
        encryption_key = force_encrypt if force_encrypt != False else encryption_key
        self.counter += 1

        meta = {"type": "value", "file": None, "direct_file": True}

        if not isinstance(key, str):
            raise TypeError("Key must be a string")
        if not isinstance(file, str):
            raise TypeError("File must be a string")

        try:
            standart_key_location = os.path.join(self.location,HASHES.get_hash(key))
            key_location =  standart_key_location if custom_key_location == "" else custom_key_location
            
            if custom_key_location != "" and not short_cut:
                self.set(key,value=key_location,file=file,compress=compress,encryption_key=encryption_key,cache_policy=cache_policy,dont_delete_cache=dont_delete_cache,dont_remove_file=dont_remove_file,short_cut = True)            

            key_location_loading = os.path.join(self.location,
                                                standart_key_location + ".l")
            random_number = random.randint(10000,99999)
            key_location_loading_indicator = os.path.join(
                self.location, standart_key_location + str(random_number) + ".li")

            key_location_reading_indicator = os.path.join(
                self.location, standart_key_location)
            key_location_compress_indicator = os.path.join(
                self.location, standart_key_location + ".co")

            if file != "":
                meta["type"] = "file"
                meta["file"] = os.path.join(
                    self.location, key_location + "." + file.split(".")[-1])
                try:
                    if not compress and encryption_key == "":
                        value = ""
                        if not dont_remove_file:
                            move(file, meta["file"])
                        else:
                            copy(file, meta["file"])
                    else:
                        meta["direct_file"] = False
                        with open(file, "rb") as f:
                            value = f.read()
                except:
                    traceback.print_exc()
                    return False

            if encryption_key != "":
                value = self.encrypt(encryption_key, value)

            the_dict = {"key": key, "value": value, "meta": meta, "short_cut": False}
            if short_cut:
                the_dict["short_cut"] = True

            if cache_policy != 0:
                the_dict["cache_time"] = time.time()
                the_dict["cache_policy"] = cache_policy
            else:
                if key in self.cache and not dont_delete_cache:
                    del self.cache[key]

            if compress:
                # create key_location_compress_indicator
                with open(key_location_compress_indicator, "wb") as f:
                    f.write(b"1")
                import mgzip

                with mgzip.open(key_location_loading, "wb") as f:
                    pickle.dump(the_dict, f)

            else:
                with open(key_location_loading, "wb") as f:
                    pickle.dump(the_dict, f)

            # Create a file that inform is loading
            with open(key_location_loading_indicator, "wb") as f:
                f.write(b"1")

            try_number = 0
            busy = True
            while not busy and try_number < 6:
                any_file = False
                for each_file in os.listdir(self.location):
                    if each_file.startswith(
                            key_location_reading_indicator) and each_file.endswith(
                                ".re"):
                        any_file = True
                if not any_file:
                    busy = False
                    break
                try_number += 1
                time.sleep(0.25)

            with contextlib.suppress(FileNotFoundError):
                move(key_location_loading, key_location)

            with contextlib.suppress(FileNotFoundError):
                os.remove(key_location_loading_indicator)

        except:
            traceback.print_exc()
            return False

        return True

    def set_withrkey(self, value) -> str:
        key = str(self.counter) + str(time.time())
        try:
            self.set(key, value)
            return key
        except:
            traceback.print_exc()
            return ""

    def transformer(self, element, encryption_key: str = ""):
        if "meta" not in element:
            element["meta"] = {"type": "value"}
        if "short_cut" not in element:
            element["short_cut"] = False
        if element["meta"]["type"] == "value":
            if encryption_key != "":
                return self.decrypt(encryption_key, element["value"])
            return element["value"]
        elif element["meta"]["type"] == "file":
            self.open_files_db.set(element["meta"]["file"], True)
            return element["meta"]["file"]


    def wait_system(self, key: str, indicator:str):
        try_number = 0
        busy = True
        while not busy and try_number < 6:
                any_file = False
                for each_file in os.listdir(self.location):
                    if each_file.startswith(
                            indicator) and each_file.endswith(
                                indicator):
                        any_file = True
                if not any_file:
                    busy = False
                    break
                try_number += 1
                time.sleep(0.25)



    def get(
        self,
        key: str,
        custom_key_location: str = "",
        encryption_key: str = "",
        no_cache: bool = False,
        raw_dict: bool = False,
        get_shotcut: bool = False,
    ):
        encryption_key = force_encrypt if force_encrypt != False else encryption_key
        if key in self.cache and not no_cache:
            cache_control = False
            currently = time.time()
            last_time = self.cache[key]["cache_time"]
            cache_policy = self.cache[key]["cache_policy"]

            if currently - last_time < cache_policy:
                cache_control = True

            if cache_control:
                return self.transformer(self.cache[key])


        standart_key_location = os.path.join(self.location,HASHES.get_hash(key))
        key_location =  standart_key_location if custom_key_location == "" else custom_key_location
            
        
        key_location_loading_indicator = os.path.join(
                self.location, standart_key_location)

        random_number = random.randint(10000,99999)
        key_location_reading_indicator = os.path.join(
                self.location, standart_key_location +str(random_number) + ".re")
        key_location_compress_indicator = os.path.join(
                self.location, standart_key_location + ".co")



        try_number = 0
        busy = True
        while not busy and try_number < 6:
                any_file = False
                for each_file in os.listdir(self.location):
                    if each_file.startswith(
                            key_location_loading_indicator) and each_file.endswith(
                                ".li"):
                        any_file = True
                if not any_file:
                    busy = False
                    break
                try_number += 1
                time.sleep(0.25)


        if not os.path.isfile(key_location):
            return None

        total_result = None
        total_result_standart = None

        try:

            # Create a file that inform is reading
            with contextlib.suppress(Exception):
                with open(key_location_reading_indicator, "wb") as f:
                    f.write(b"1")

            # Read the file
            if os.path.exists(key_location_compress_indicator):
                import mgzip
                with mgzip.open(key_location,
                                "rb") as f:
                    result = pickle.load(f)
            else:
                with open(key_location,
                          "rb") as f:
                    result = pickle.load(f)

            # Transform the result
            total_result_standart = result
            try:
                total_result = self.transformer(result, encryption_key=encryption_key)
            except TypeError:
                traceback.print_exc()
                total_result = result            

            # Add to cache
            if "cache_time" in total_result_standart:
                self.cache[key] = total_result_standart

        except EOFError or FileNotFoundError:
            traceback.print_exc()

        # Delete the file that inform is reading
        if os.path.isfile(key_location_reading_indicator):
            with contextlib.suppress(Exception):
                os.remove(key_location_reading_indicator)

        # Create the file if there is an compress or encryption situation
        if total_result_standart["meta"]["type"] == "file":
            if not total_result_standart["meta"]["direct_file"]:
                with open(total_result_standart["meta"]["file"], "wb") as f:
                    the_bytes = total_result_standart["value"]
                    if encryption_key != "":
                        the_bytes = self.decrypt(encryption_key, the_bytes)
                    f.write(the_bytes)

        # Return the result
        if raw_dict:

            if total_result_standart["short_cut"] and not get_shotcut:
                total_result_standart = self.get(key, custom_key_location=total_result_standart["value"],
                                    encryption_key=encryption_key,
                                    no_cache=no_cache,
                                    raw_dict=raw_dict)


            return total_result_standart



        if total_result_standart["short_cut"] and not get_shotcut:
            total_result = self.get(key, custom_key_location=total_result_standart["value"],
                                    encryption_key=encryption_key,
                                    no_cache=no_cache,
                                    raw_dict=raw_dict)


        return total_result

    def get_key(self, key_location: str):
        key_location_compress_indicator = os.path.join(self.location,
                                                       key_location + ".co")
        if not os.path.isfile(os.path.join(self.location, key_location)):
            return None
        total_result = None

        try:
            if os.path.exists(key_location_compress_indicator):
                import mgzip
                with mgzip.open(os.path.join(self.location, key_location),
                                "rb") as f:
                    result = pickle.load(f)
            else:
                with open(os.path.join(self.location, key_location),
                          "rb") as f:
                    result = pickle.load(f)


            if not "cache_time" in result:
                result["cache_time"] = 0
            if not "cache_policy" in result:
                result["cache_policy"] = 0
            if not "meta" in result:
                result["meta"] = {"type": "value"}
            try:
                total_result = result["key"]
            except TypeError:
                total_result = False


        except EOFError or FileNotFoundError:
            pass
        return total_result

    def get_all(self):
        return self.dict()


    def get_count(self):
        return len(self.dict(no_data=True))

    def delete(self, key: str) -> bool:
        try:
            if key in self.cache:
                del self.cache[key]
            key_location = os.path.join(self.location,
                                        HASHES.get_hash(key))
            key_location_compress_indicator = os.path.join(
                self.location, key_location + ".co")

            with contextlib.suppress(TypeError):
                maybe_file = self.get(key)
                if os.path.exists(maybe_file):
                    os.remove(maybe_file)

            the_get = self.get(key, no_cache=True, raw_dict=True, get_shotcut=True)
            with contextlib.suppress(TypeError):
                    maybe_file = self.get(key, custom_key_location=the_get["value"])
                    if os.path.exists(maybe_file):
                        os.remove(maybe_file)   
            with contextlib.suppress(TypeError):
                if the_get["short_cut"]:
                    os.remove(the_get["value"])             

            if os.path.exists(key_location_compress_indicator):
                os.remove(
                    os.path.join(self.location,
                                 key_location_compress_indicator))
            if os.path.exists(os.path.join(self.location, key_location)):
                os.remove(os.path.join(self.location, key_location))
        except:
            traceback.print_exc()
            return False

        return True

    def delete_all(self) -> bool:
        try:
            for key in self.dict():
                self.delete(key)
        except:
            traceback.print_exc()
            return False
        return True

    def dict(self, encryption_key: str = "", no_data: bool = False):
        encryption_key = force_encrypt if force_encrypt != False else encryption_key
        result = {}
        for key in os.listdir(self.location):
            if not "." in key:
                with contextlib.suppress(Exception):
                    the_key = self.get_key(key)
                    if not the_key is None:
                        if the_key != False:
                            result_of_key = (self.get(
                                the_key, encryption_key=encryption_key)
                                            if not no_data else True)
                            if not result_of_key is None:
                                result[the_key] = result_of_key
        return result

    def size_all(self) -> int:
        # Calculate self.location size
        total_size = 0

        for dirpath, dirnames, filenames in os.walk(self.location):
            for f in filenames:
                fp = os.path.join(dirpath, f)
                total_size += os.path.getsize(fp)

        return total_size

    def size(self, key: str) -> int:
        total_size = 0
        try:
            key_location = os.path.join(self.location,
                                        HASHES.get_hash(key))

            key_location_compress_indicator = os.path.join(
                self.location, key_location + ".co")

            if os.path.exists(key_location_compress_indicator):
                total_size += os.path.getsize(
                    os.path.join(self.location, key_location + ".co"))

            with contextlib.suppress(TypeError):
                maybe_file = self.get(key)
                if os.path.exists(maybe_file):
                    total_size += os.path.getsize(maybe_file)

            total_size += os.path.getsize(
                os.path.join(self.location, key_location))
        except:
            traceback.print_exc()

        return total_size

    def backup(self, backup_location: str) -> str:
        # create a name for backup that a date
        name = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
        location = os.path.join(backup_location, name)
        make_archive(location, "zip", self.location)
        move(location + ".zip", location + ".KOT")
        return location + ".KOT"

    def restore(self, backup_location: str) -> bool:
        try:
            move(backup_location, backup_location.replace(".KOT", ".zip"))
            backup_location = backup_location.replace(".KOT", ".zip")
            unpack_archive(backup_location, self.location)
            move(backup_location, backup_location.replace(".zip", ".KOT"))
            return True
        except:
            traceback.print_exc()
            return False





def main():
    import fire

    fire.Fire(KOT)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from hashlib import sha256

from naruno.lib.encryption import decrypt
from naruno.lib.settings_system import the_settings
from naruno.wallet.get_saved_wallet import get_saved_wallet
from naruno.wallet.wallet_create import wallet_create


def wallet_import(wallet, mode, password=None):
    """
    A function for get info about a wallet.

    Inputs:
      * account: Account index of saved accounts (if you give -1 the default wallet will use)
      * mode: Information mode [0 = Public key | 1 = Private key (needs password) | 2 = Returns sha256 of password | 3 = Returns the address of account]
      * password: Some function needed password for operation you can give with this input
    """

    account = wallet

    temp_saved_wallet = get_saved_wallet()

    number_of_wallet = len(temp_saved_wallet)
    if not number_of_wallet:
        wallet_create("123")
        temp_saved_wallet = get_saved_wallet()

    if isinstance(wallet, int):
        if wallet == -1:
            account = list(temp_saved_wallet)[the_settings()["wallet"]]

        elif wallet <= (len(temp_saved_wallet) - 1):
            account = list(temp_saved_wallet)[account]
        else:
            return False
    if mode == 0:
        return temp_saved_wallet[account]["publickey"]
    elif mode == 1:
        if password is None:
            if list(temp_saved_wallet).index(account) != 0:
                return False
            my_private_key = temp_saved_wallet[account]["privatekey"]
            return my_private_key

        elif list(temp_saved_wallet).index(account) != 0:
            return decrypt(temp_saved_wallet[account]["privatekey"], password)
        else:
            if wallet != -1:
                return False
            my_private_key = temp_saved_wallet[account]["privatekey"]
            return my_private_key
    elif mode == 2:
        return temp_saved_wallet[account]["password_sha256"]

    elif mode == 3:
        my_address = temp_saved_wallet[account]["publickey"]
        my_address = "".join([
            l.strip() for l in my_address.splitlines()
            if l and not l.startswith("-----")
        ])
        my_address = Address(my_address)
        return my_address
    else:
        return False


def wallet_import_all(mode, password=None):
    temp_saved_wallet = get_saved_wallet()
    the_result = []

    for account in temp_saved_wallet:
        the_result.append(wallet_import(account, mode, password))
    return the_result


def Address(publickey):
    the_public_key = "".join([
        l.strip() for l in publickey.splitlines()
        if l and not l.startswith("-----")
    ])
    return sha256(
        sha256(the_public_key.encode("utf-8")).hexdigest().encode(
            "utf-8")).hexdigest()[-40:]



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import copy
import threading
import traceback
from naruno.blockchain.block.block_main import Block
from naruno.consensus.sync.send_block import send_block
from naruno.consensus.sync.send_block_hash import send_block_hash
from naruno.lib.log import get_logger
from naruno.node.server.server import server
import naruno
logger = get_logger("CONSENSUS")


sync_round_1 = True
sync_round_2 = False

sended_txs = []

def sync(
    block: Block,
    pending_list_txs: list = None,
    custom_server: server = None,
    send_block_error: bool = False,
    send_block_hash_error: bool = False,
    send_transaction_error: bool = False,
):
    """
    Data sending consists of 3 stages.
    Block sending,blockhash sending and transection.
    It shares the data of the existing chains with the nodes.
    """

    logger.info("Data sending process is starting")
    the_server = server.Server if custom_server is None else custom_server

    if naruno.consensus.sync.sync.sync_round_1:
        threading.Thread(
            target=send_block,
            args=(block, the_server, send_block_error),
        ).start()
        naruno.consensus.sync.sync.sync_round_1 = False
  
    if naruno.consensus.sync.sync.sync_round_2:
            threading.Thread(
                target=send_block_hash,
                args=(block, the_server, send_block_hash_error),
            ).start()
            naruno.consensus.sync.sync.sync_round_2 = False

    logger.debug("Transactions is sending to the unl nodes")
    the_transactions_list = copy.copy(block.validating_list)
    if pending_list_txs is not None:
        the_transactions_list += pending_list_txs
    for i in the_transactions_list:
        try:
            if not i.signature in naruno.consensus.sync.sync.sended_txs:
                naruno.consensus.sync.sync.sended_txs.append(i.signature)
                the_server.send_transaction(i)
            if send_transaction_error:
                raise Exception("Transaction sending error")
        except Exception as e:
            traceback.print_exc()
            logger.error(f"Transaction sending error: {e}")



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import os
import sqlite3

from naruno.config import TEMP_ACCOUNTS_PATH
from naruno.lib.config_system import get_config
from naruno.lib.kot import KOT
from naruno.lib.log import get_logger
import hashlib

accounts_db = KOT("accounts_db", folder=get_config()["main_folder"] + "/db")

accounts_ram_db = {}


operations_hashes = []
logger = get_logger("ACCOUNTS")

def get_ram_accounts(name: str, reset: bool = False):
    if not name in accounts_ram_db or reset:
        record = accounts_db.get("accounts", custom_key_location=name)

        if record is None:
            accounts_ram_db[name] = {}
        else:
            accounts_ram_db[name] = record
    return name


def SaveAccounts(new_account,
                 custom_TEMP_ACCOUNTS_PATH=None,
                 reset: bool = False,sequence=None):
    """
    Saves the accounts to the TEMP_ACCOUNTS_PATH.
    """

    os.chdir(get_config()["main_folder"])

    if type(new_account) != list:
        new_account = [new_account]

    if sequence is not None:
        logger.info("Sequence is not node")
        the_string=str(sequence)+str(custom_TEMP_ACCOUNTS_PATH)
        logger.debug(f"String: {the_string}")
        the_hash = hashlib.sha256(the_string.encode()).hexdigest()
        logger.debug(f"Hash: {the_hash}")
        if the_hash in operations_hashes:
            logger.warning("Two times try")
            return
        else:
            logger.info("Passed")
            operations_hashes.append(the_hash)



    the_TEMP_ACCOUNTS_PATH = (TEMP_ACCOUNTS_PATH if custom_TEMP_ACCOUNTS_PATH
                              is None else custom_TEMP_ACCOUNTS_PATH)

    the_TEMP_ACCOUNTS_PATH = os.path.join(get_config()["main_folder"],
                                          the_TEMP_ACCOUNTS_PATH)

    ram_db_record = get_ram_accounts(the_TEMP_ACCOUNTS_PATH)
    if reset:
        accounts_ram_db[ram_db_record] = {}
    for account in new_account:
        accounts_ram_db[ram_db_record][account.Address] = [
            account.sequence_number,
            account.balance,
        ]

    accounts_db.set(
        "accounts",
        accounts_ram_db[ram_db_record],
        custom_key_location=the_TEMP_ACCOUNTS_PATH,
    )


    return accounts_ram_db[ram_db_record]


#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import json
import os

from naruno.config import CONFIG_PATH
from naruno.lib.kot import KOT

config_db = KOT("config",
                folder=os.path.join(os.path.dirname(__file__), "..") + "/db")


def save_config(config):
    """
    Saves the settings.
    """

    config_db.set("config", config)


def create_and_save_the_configs():
    """
    Creates and saves configs.
    """

    temp_json = {"main_folder": os.path.join(os.path.dirname(__file__), "..")}
    save_config(temp_json)
    return temp_json


def get_config():
    """
    Returns the configs. If it doesn't exist, it creates,
    saves and returns.
    """

    record = config_db.get("config")
    if record is None:
        return create_and_save_the_configs()
    return record



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import os
import sqlite3

import naruno
from naruno.accounts.save_accounts import get_ram_accounts
from naruno.config import TEMP_ACCOUNTS_PATH
from naruno.lib.config_system import get_config


def GetAccounts(custom_TEMP_ACCOUNTS_PATH=None, reset: bool = False):
    """
    Returns the accounts from TEMP_ACCOUNTS_PATH.
    """
    os.chdir(get_config()["main_folder"])

    the_TEMP_ACCOUNTS_PATH = (TEMP_ACCOUNTS_PATH if custom_TEMP_ACCOUNTS_PATH
                              is None else custom_TEMP_ACCOUNTS_PATH)
    the_TEMP_ACCOUNTS_PATH = os.path.join(get_config()["main_folder"],
                                          the_TEMP_ACCOUNTS_PATH)
    return naruno.accounts.save_accounts.accounts_ram_db[get_ram_accounts(
        the_TEMP_ACCOUNTS_PATH, reset=reset)]



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import json
import os
import sqlite3

from naruno.config import TEMP_BLOCKSHASH_PART_PATH
from naruno.config import TEMP_BLOCKSHASH_PATH
from naruno.lib.config_system import get_config
from naruno.lib.kot import KOT
from naruno.lib.log import get_logger

blockshash_db = KOT("blockshash", folder=get_config()["main_folder"] + "/db")

logger = get_logger("BLOCKCHAIN")


def SaveBlockshash(the_blockshash, custom_TEMP_BLOCKSHASH_PATH=None):
    """
    Saves the blockshash to the TEMP_BLOCKSHASH_PATH.
    """
    os.chdir(get_config()["main_folder"])

    the_TEMP_BLOCKSHASH_PATH = (TEMP_BLOCKSHASH_PATH
                                if custom_TEMP_BLOCKSHASH_PATH is None else
                                custom_TEMP_BLOCKSHASH_PATH)

    the_TEMP_BLOCKSHASH_PATH = os.path.join(get_config()["main_folder"],
                                            the_TEMP_BLOCKSHASH_PATH)

    if type(the_blockshash) != list:
        the_blockshash = [the_blockshash]

    the_blockshash = (GetBlockshash(
        custom_TEMP_BLOCKSHASH_PATH=custom_TEMP_BLOCKSHASH_PATH) +
                      the_blockshash)

    blockshash_db.set("blockshash",
                      the_blockshash,
                      custom_key_location=the_TEMP_BLOCKSHASH_PATH)


def SaveBlockshash_part(the_blockshash, custom_TEMP_BLOCKSHASH_PART_PATH=None):
    """
    Saves the blockshash part to the TEMP_BLOCKSHASH_PART_PATH.
    """
    os.chdir(get_config()["main_folder"])

    the_TEMP_BLOCKSHASH_PART_PATH = (TEMP_BLOCKSHASH_PART_PATH if
                                     custom_TEMP_BLOCKSHASH_PART_PATH is None
                                     else custom_TEMP_BLOCKSHASH_PART_PATH)

    the_TEMP_BLOCKSHASH_PART_PATH = os.path.join(
        get_config()["main_folder"], the_TEMP_BLOCKSHASH_PART_PATH)

    if type(the_blockshash) != list:
        the_blockshash = [the_blockshash]

    the_blockshash = (GetBlockshash_part(
        custom_TEMP_BLOCKSHASH_PART_PATH=custom_TEMP_BLOCKSHASH_PART_PATH) +
                      the_blockshash)

    blockshash_db.set(
        "blockshash_part",
        the_blockshash,
        custom_key_location=the_TEMP_BLOCKSHASH_PART_PATH,
    )


def GetBlockshash(custom_TEMP_BLOCKSHASH_PATH=None):
    """
    Returns the blockshash.
    """
    os.chdir(get_config()["main_folder"])

    the_TEMP_BLOCKSHASH_PATH = (TEMP_BLOCKSHASH_PATH
                                if custom_TEMP_BLOCKSHASH_PATH is None else
                                custom_TEMP_BLOCKSHASH_PATH)

    the_TEMP_BLOCKSHASH_PATH = os.path.join(get_config()["main_folder"],
                                            the_TEMP_BLOCKSHASH_PATH)

    record = blockshash_db.get("blockshash",
                               custom_key_location=the_TEMP_BLOCKSHASH_PATH)

    return record if record is not None else []


def GetBlockshash_part(custom_TEMP_BLOCKSHASH_PART_PATH=None):
    """
    Returns the blockshash part.
    """
    os.chdir(get_config()["main_folder"])

    the_TEMP_BLOCKSHASH_PART_PATH = (TEMP_BLOCKSHASH_PART_PATH if
                                     custom_TEMP_BLOCKSHASH_PART_PATH is None
                                     else custom_TEMP_BLOCKSHASH_PART_PATH)

    the_TEMP_BLOCKSHASH_PART_PATH = os.path.join(
        get_config()["main_folder"], the_TEMP_BLOCKSHASH_PART_PATH)

    record = blockshash_db.get(
        "blockshash_part", custom_key_location=the_TEMP_BLOCKSHASH_PART_PATH)

    return record if record is not None else []



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import contextlib
import copy
import json
import os
import random
import socket
import time
from hashlib import sha256
from shutil import move
from threading import Thread
import traceback

from naruno.accounts.save_accounts import accounts_db
from naruno.blockchain.block.block_main import Block
from naruno.blockchain.block.blocks_hash import blockshash_db
from naruno.blockchain.block.change_transaction_fee import ChangeTransactionFee
from naruno.blockchain.block.get_block import GetBlock
from naruno.blockchain.block.save_block import SaveBlock
from naruno.config import CONNECTED_NODES_PATH
from naruno.config import LOADING_ACCOUNTS_PATH
from naruno.config import LOADING_BLOCK_PATH
from naruno.config import LOADING_BLOCKSHASH_PART_PATH
from naruno.config import LOADING_BLOCKSHASH_PATH
from naruno.config import PENDING_TRANSACTIONS_PATH
from naruno.config import TEMP_ACCOUNTS_PATH
from naruno.config import TEMP_BLOCK_PATH
from naruno.config import TEMP_BLOCKSHASH_PART_PATH
from naruno.config import TEMP_BLOCKSHASH_PATH
from naruno.lib.config_system import get_config
from naruno.lib.kot import KOT
from naruno.lib.log import get_logger
from naruno.node.client.client import client
from naruno.node.unl import Unl
from naruno.transactions.get_transaction import GetTransaction
from naruno.transactions.transaction import Transaction
from naruno.wallet.ellipticcurve.ecdsa import Ecdsa
from naruno.wallet.ellipticcurve.privateKey import PrivateKey
from naruno.wallet.ellipticcurve.publicKey import PublicKey
from naruno.wallet.ellipticcurve.signature import Signature
from naruno.wallet.wallet_import import wallet_import

from naruno.node.get_candidate_blocks import self_candidates
import naruno

connectednodes_db = KOT("connectednodes",
                        folder=get_config()["main_folder"] + "/db")
 


tx_signature_list = {}



class server(Thread):
    Server = None
    id = wallet_import(0, 0)

    def __init__(
        self,
        host,
        port,
        save_messages=False,
        test=False,
        custom_variables=False,
        custom_TEMP_BLOCK_PATH=None,
        custom_TEMP_ACCOUNTS_PATH=None,
        custom_TEMP_BLOCKSHASH_PATH=None,
        custom_TEMP_BLOCKSHASH_PART_PATH=None,
        custom_LOADING_BLOCK_PATH=None,
        custom_LOADING_ACCOUNTS_PATH=None,
        custom_LOADING_BLOCKSHASH_PATH=None,
        custom_LOADING_BLOCKSHASH_PART_PATH=None,
        custom_CONNECTED_NODES_PATH=None,
        custom_PENDING_TRANSACTIONS_PATH=None,
        time_control=None,
        custom_id=None,
    ):
        Thread.__init__(self)
        self.running = True

        self.host = host
        self.port = port
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.bind((self.host, self.port))
        self.sock.listen(1)

        self.clients = []
        self.sync_clients = []

        self.messages = []
        self.our_messages = []
        self.save_messages = save_messages

        self.TEMP_BLOCK_PATH = (TEMP_BLOCK_PATH if custom_TEMP_BLOCK_PATH
                                is None else custom_TEMP_BLOCK_PATH)
        self.TEMP_ACCOUNTS_PATH = (TEMP_ACCOUNTS_PATH
                                   if custom_TEMP_ACCOUNTS_PATH is None else
                                   custom_TEMP_ACCOUNTS_PATH)
        self.TEMP_BLOCKSHASH_PATH = (TEMP_BLOCKSHASH_PATH
                                     if custom_TEMP_BLOCKSHASH_PATH is None
                                     else custom_TEMP_BLOCKSHASH_PATH)
        self.TEMP_BLOCKSHASH_PART_PATH = (
            TEMP_BLOCKSHASH_PART_PATH if custom_TEMP_BLOCKSHASH_PART_PATH
            is None else custom_TEMP_BLOCKSHASH_PART_PATH)
        self.LOADING_BLOCK_PATH = (LOADING_BLOCK_PATH
                                   if custom_LOADING_BLOCK_PATH is None else
                                   custom_LOADING_BLOCK_PATH)
        self.LOADING_ACCOUNTS_PATH = (LOADING_ACCOUNTS_PATH
                                      if custom_LOADING_ACCOUNTS_PATH is None
                                      else custom_LOADING_ACCOUNTS_PATH)
        self.LOADING_BLOCKSHASH_PATH = (LOADING_BLOCKSHASH_PATH if
                                        custom_LOADING_BLOCKSHASH_PATH is None
                                        else custom_LOADING_BLOCKSHASH_PATH)
        self.LOADING_BLOCKSHASH_PART_PATH = (
            LOADING_BLOCKSHASH_PART_PATH if custom_LOADING_BLOCKSHASH_PART_PATH
            is None else custom_LOADING_BLOCKSHASH_PART_PATH)

        self.CONNECTED_NODES_PATH = (None if custom_CONNECTED_NODES_PATH
                                     is None else custom_CONNECTED_NODES_PATH)

        self.PENDING_TRANSACTIONS_PATH = (
            PENDING_TRANSACTIONS_PATH if custom_PENDING_TRANSACTIONS_PATH
            is None else custom_PENDING_TRANSACTIONS_PATH)

        self.custom_variables = custom_variables

        self.time_control = 10 if time_control is None else time_control

        if custom_id is not None:
            self.id = custom_id
        else:
            self.id = server.id

        self.logger = get_logger(f"NODE_{self.host}_{self.port}")

        self.logger.info(f"Server started as {server.id}")

        self.logger.debug("Save messages: " + str(save_messages))
        self.logger.debug("Test mode: " + str(test))


        self.send_busy = []


        a_block = Block("onur")
        self.logger.debug(f"Block max_data_size: {a_block.max_data_size}")
        self.logger.debug(f"Block max_tx_number: {a_block.max_tx_number}")
        self.buffer_size = 6525 + int(
            (a_block.max_data_size // a_block.max_tx_number) * 1.5)
        self.buffer_size_2 = 6525 + int(
            (a_block.max_data_size // a_block.max_tx_number) * 1.5)
        self.buffer_size_3 = 6525 + int(
            (a_block.max_data_size // a_block.max_tx_number) * 5)    
        self.buffer_size_4 = 6525 + int(
            (a_block.max_data_size // a_block.max_tx_number) * 0.5)   

      
        if not test:
            self.__class__.Server = self
            self.start()

    def check_connected(self, host, port):
        for a_client in self.clients:
            if a_client.host == host and a_client.port == port:
                return True

        return False

    def run(self):
        self.logger.info("Server ear started")
        self.sock.settimeout(10.0)
        while self.running:
            with contextlib.suppress(Exception):
                conn, addr = self.sock.accept()
                self.logger.debug(f"New connection request: {addr}")
                data = conn.recv(1024)
                
                raw_id = data.decode("utf-8")
                client_id = raw_id.split("-")[0]
                client_type = int(raw_id.split("-")[1])
                conn.send((self.id+"-"+str(client_type)).encode("utf-8"))
                self.logger.debug(f"New connection id: {client_id}")
                if Unl.node_is_unl(client_id):
                    self.logger.info(f"Confirmed")
                    self.clients.append(client(conn, addr, client_id, self, c_type=client_type))
                    self.save_connected_node(addr[0], addr[1], client_id)
                else:
                    self.logger.warning(f"Rejected")
            time.sleep(0.01)

    def stop(self):
        self.running = False
        socket.socket(socket.AF_INET, socket.SOCK_STREAM).connect(
            (self.host, self.port))
        for c in self.clients:
            c.stop()
        time.sleep(1)
        for c in self.clients:
            c.join()
        self.sock.close()

    def prepare_message(self, data):
        data["id"] = server.id
        data["timestamp"] = str(time.time())
        sign = Ecdsa.sign(
            str(data),
            PrivateKey.fromPem(wallet_import(0, 1)),
        ).toBase64()

        data["signature"] = sign
        return data

    def send(self, data, except_client=None, c_type=0):
        data = self.prepare_message(data)

        for a_client in self.clients:
            if a_client != except_client:
                if a_client.c_type == c_type:
                    self.send_client(a_client, data, ready_to_send=True, c_type=c_type)
        try:
            del data["buffer"]
        except KeyError:
            pass
        return data

    def send_client(self, node, data, ready_to_send=False, c_type=0):
        self.logger.debug(
            f"Sending message: {data} to {node.host}:{node.port}={node.id}")
        if not ready_to_send:
            data = self.prepare_message(data)



        if c_type == 0:
            the_buffer = self.buffer_size
        elif c_type == 1:
            the_buffer = self.buffer_size_2
        elif c_type == 2:
            the_buffer = self.buffer_size_3
        elif c_type == 3:
            the_buffer = self.buffer_size_4

        before_buffer_size = len(json.dumps(data).encode("utf-8"))
        self.logger.debug(
            f"Before buffer size: {before_buffer_size}")

        if before_buffer_size < the_buffer:
            self.logger.debug("Buffer is not full")
            self.logger.debug(f"the_buffer: {the_buffer}")
            self.logger.debug(f"before_buffer_size: {before_buffer_size}")
            result = ((the_buffer - before_buffer_size) - 14)
            self.logger.debug(f"result: {result}")
            data["buffer"] = "0" * result
            self.logger.debug(
                f"After buffer size: {len(json.dumps(data).encode('utf-8'))}")
        while node.id+str(c_type) in self.send_busy:
            time.sleep(0.01)
        self.send_busy.append(node.id+str(c_type))
        try:
            with contextlib.suppress(socket.timeout):
                node.socket.sendall(json.dumps(data).encode("utf-8"))
                time.sleep(0.02)
        except:
            traceback.print_exc()
        self.send_busy.remove(node.id+str(c_type))
        with contextlib.suppress(KeyError):
            del data["buffer"]
        if self.save_messages:
            self.our_messages.append(data)
        return data

    def get_message(self, client, data, hash_of_data=""):
        self.logger.info(
            f"Starting to proccess the message ({hash_of_data}) of {client.id}:{client.host}:{client.port}"
        )
        if self.check_message(data):
            self.logger.debug(f"Message is valid")
            if self.save_messages:
                self.messages.append(data)
            self.direct_message(client, data, hash_of_data)
        else:
            self.logger.error(f"Message is not in a valid format")

    def check_message(self, data):
        if "id" not in data:
            self.logger.error("No id")
            return False
        if "signature" not in data:
            self.logger.error("No signature")
            return False
        if "timestamp" not in data:
            self.logger.error("No timestamp")
            return False
        # the_control = time.time() - float(data["timestamp"])
        # if the_control > self.time_control:
        #    logger.debug("Time control is not true")
        #    return False
        # remove sign from data
        sign = data["signature"]
        del data["signature"]
        message = str(data)
        data["signature"] = sign

        signature_verify = Ecdsa.verify(
            message,
            Signature.fromBase64(sign),
            PublicKey.fromPem(data["id"]),
        )

        if not signature_verify:
            self.logger.error("Signature not valid")
            return False

        return True


    def _connect(self, host, port, c_type=0):
            self.logger.info(
                "New connection request confirmed trying to connect")
            conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            conn.settimeout(10.0)
            addr = (host, port)
            try:
                conn.connect(addr)
                conn.send((self.id+"-"+str(c_type)).encode("utf-8"))                
                raw_id = conn.recv(1024).decode("utf-8")
                client_id = raw_id.split("-")[0]
                client_type = int(raw_id.split("-")[1])                
                if Unl.node_is_unl(client_id):
                    self.logger.info(
                        f"Succesfully connected to {client_id} on {host}:{port}"
                    )
                    self.clients.append(client(conn, addr, client_id, self, c_type=client_type))
                    self.save_connected_node(addr[0], addr[1], client_id)
                    return True
            except socket.timeout:
                self.logger.warning(f"Connection timeout")
                conn.close()
            except ConnectionRefusedError:
                self.logger.warning(f"Connection refused")
                conn.close()    

    def connect(self, host, port):
        self.logger.info(f"Asking for new node on {host}:{port}")
        connected = self.check_connected(host=host, port=port)
        if not connected:
            self._connect(host, port)
            time.sleep(1)
            self._connect(host, port, c_type=1)
            time.sleep(1)
            self._connect(host, port, c_type=2)
            time.sleep(1)
            self._connect(host, port, c_type=3)
        else:
            self.logger.warning("Already connected")

    @staticmethod
    def get_connected_nodes(custom_CONNECTED_NODES_PATH=None):
        """
        Returns the connected nodes.
        """

        the_pending_list = {}
        all_records = (connectednodes_db.get_all()
                       if custom_CONNECTED_NODES_PATH is None else KOT(
                           "connectednodes" + custom_CONNECTED_NODES_PATH,
                           folder=get_config()["main_folder"] + "/db",
                       ).get_all())
        for entry in all_records:
            loaded_json = all_records[entry]
            the_pending_list[loaded_json["host"] + str(loaded_json["port"]) +
                             loaded_json["id"]] = loaded_json

        return the_pending_list

    def save_connected_node(self, host, port, node_id):
        """
        Saves the connected nodes.
        """

        node_list = {}
        node_list["id"] = node_id
        node_list["host"] = host
        node_list["port"] = port

        node_id = sha256(
            (node_id + host + str(port)).encode("utf-8")).hexdigest()

        connectednodes_db.set(
            node_id, node_list) if self.CONNECTED_NODES_PATH is None else KOT(
                "connectednodes" + self.CONNECTED_NODES_PATH,
                folder=get_config()["main_folder"] + "/db",
            ).set(node_id, node_list)

    @staticmethod
    def connectionfrommixdb(custom_server=None,
                            custom_CONNECTED_NODES_PATH=None):
        """
        Connects to the mixdb.
        """
        the_server = server.Server if custom_server is None else custom_server
        the_CONNECTED_NODES_PATH = (the_server.CONNECTED_NODES_PATH
                                    if custom_CONNECTED_NODES_PATH is None else
                                    custom_CONNECTED_NODES_PATH)
        node_list = the_server.get_connected_nodes(
            custom_CONNECTED_NODES_PATH=the_CONNECTED_NODES_PATH)
        for element in node_list:
            with contextlib.suppress(Exception):
                the_server.connect(
                    node_list[element]["host"],
                    node_list[element]["port"],
                )

    def connected_node_delete(self, node):
        """
        Deletes a connected node.
        """

        node_id = sha256((node["id"] + node["host"] +
                          str(node["port"])).encode("utf-8")).hexdigest()
        connectednodes_db.delete(
            node_id) if self.CONNECTED_NODES_PATH is None else KOT(
                "connectednodes" + self.CONNECTED_NODES_PATH,
                folder=get_config()["main_folder"] + "/db",
            ).delete(node_id)

    def direct_message(self, node, data, hash_of_data):
        if "sendmefullblock" == data["action"]:
            self.send_block_to_other_nodes(node, hash_of_data=hash_of_data)

        if "fullblock" == data["action"]:
            self.get_full_chain(data, node, hash_of_data=hash_of_data)

        if "fullaccounts" == data["action"]:
            self.get_full_accounts(data, node, hash_of_data=hash_of_data)

        if "fullblockshash" == data["action"]:
            self.get_full_blockshash(data, node, hash_of_data=hash_of_data)

        if "fullblockshash_part" == data["action"]:
            self.get_full_blockshash_part(data,
                                          node,
                                          hash_of_data=hash_of_data)

        if "transactionrequest" == data["action"]:
            self.get_transaction(data, node, hash_of_data=hash_of_data)

        if "myblock" == data["action"]:
            self.get_candidate_block(data, node, hash_of_data=hash_of_data)

        if "myblockhash" == data["action"]:
            self.get_candidate_block_hash(data,
                                          node,
                                          hash_of_data=hash_of_data)

    def send_me_full_block(self, node=None):
        the_node = node
        choices = []
        if the_node is None:
            for i in self.clients:
                if i.c_type == 2:
                    choices.append(i)
            the_node = random.choice(choices)

        self.logger.info(
            f"Sending sendmefullblock to {the_node.id}:{the_node.host}:{the_node.port}"
        )
        self.send_client(the_node, {"action": "sendmefullblock"}, c_type=2)

    def send_my_block(self, block: Block):
        self.logger.info(f"Sending my block to all nodes")
        system = self_candidates(block)[0]

        

        new_list = []

        signature_list = []

        for element in system.validating_list:
            tx_json = element.dump_json()
            new_list.append(tx_json)
            signature_list.append(element.signature)

        first_element = [new_list[0]] if len(new_list) > 0 else []

        data = {
            "action": "myblock",
            "transaction": first_element,
            "total_length": len(new_list),
            "sequence_number":
            system.sequence_number+system.empty_block_number,
            "adding": False,
        }

        self.send(data)
        time.sleep(2)

        if len(new_list) > 1:
            for element in new_list[1:]:
                data = {
                    "action": "myblock",
                    "transaction": [element],
                    "total_length": len(new_list),
                    "sequence_number":
                    system.sequence_number+system.empty_block_number,
                    "adding": True,
                }

                self.send(data)


    def send_my_block_hash(self, block):
        self.logger.info(f"Sending my block hash to all nodes")
         
        system = self_candidates(block)[1]

        data = {
            "action": "myblockhash",
            "hash": system.hash,
            "previous_hash": system.previous_hash,
            "sequence_number":
            system.sequence_number+system.empty_block_number,
        }

        self.send(data, c_type=3)

    def get_candidate_block(self, data, node: client, hash_of_data=""):
        self.logger.info(f"Getting candidate block with {hash_of_data}")
        self.logger.debug(
            f"Getting candidate block from {node.id}:{node.host}:{node.port}")
        if node.candidate_block is None:
            node.candidate_block = data
            return
        if data["sequence_number"] > node.candidate_block["sequence_number"]:
            self.logger.debug("Candidate block is updated directly")
            if len(node.candidate_block_history) >= 5:
                node.candidate_block_history.pop(0)

            node.candidate_block_history.append(copy.copy(
                node.candidate_block))
            node.candidate_block = data
        else:
            self.logger.debug("Candidate block is not updated directly")
            if node.candidate_block["total_length"] <= data["total_length"]:
                if node.candidate_block["total_length"] == data[
                        "total_length"]:
                    if data["adding"]:
                        control = True
                        for i in node.candidate_block["transaction"]:
                            if i["signature"] == data["transaction"][0]["signature"]:
                                control = False
                        if control:
                            node.candidate_block["transaction"].append(data["transaction"][0])

                else:
                    node.candidate_block = data

    def get_candidate_block_hash(self, data, node: client, hash_of_data=""):
        self.logger.info(f"Getting candidate block hash with {hash_of_data}")
        if node.candidate_block_hash is None:
            node.candidate_block_hash = data
            return

        data["sender"] = node.id

        if data["sequence_number"] > node.candidate_block_hash[
                "sequence_number"]:
            if len(node.candidate_block_hash_history) >= 5:
                node.candidate_block_hash_history.pop(0)

            node.candidate_block_hash_history.append(
                copy.copy(node.candidate_block_hash))
            node.candidate_block_hash = data
        else:
            #if len(node.candidate_block_hash["hash"]) <= len(data["hash"]):
            node.candidate_block_hash = data

    def send_full_chain(self, node=None):
        log_text = ("Sending full chain" if node is None else
                    f"Sending full chain to {node.id}:{node.host}:{node.port}")
        self.logger.info(log_text)
        file = open(self.TEMP_BLOCK_PATH, "rb")
        SendData = file.read(1024)
        while SendData:
            data = {
                "action": "fullblock",
                "byte": (SendData.decode(encoding="iso-8859-1")),
            }
            if node is None:
                self.send(data, c_type=2)
            else:
                self.send_client(node, data, c_type=2)

            SendData = file.read(1024)

            if not SendData:
                data = {
                    "action": "fullblock",
                    "byte": "end",
                }
                if node is None:
                    self.send(data, c_type=2)
                else:
                    self.send_client(node, data, c_type=2)

    def send_full_accounts(self, node=None):
        the_TEMP_ACCOUNTS_PATH = self.TEMP_ACCOUNTS_PATH
        file = open(the_TEMP_ACCOUNTS_PATH, "rb")
        SendData = file.read(1024)
        while SendData:
            data = {
                "action": "fullaccounts",
                "byte": (SendData.decode(encoding="iso-8859-1")),
            }

            if node is None:
                self.send(data, c_type=2)
            else:
                self.send_client(node, data, c_type=2)

            SendData = file.read(1024)
            if not SendData:
                data = {"action": "fullaccounts", "byte": "end"}
                if node is None:
                    self.send(data, c_type=2)
                else:
                    self.send_client(node, data, c_type=2)

    def send_full_blockshash(self, node=None):
        the_TEMP_BLOCKSHASH_PATH = self.TEMP_BLOCKSHASH_PATH
        file = open(the_TEMP_BLOCKSHASH_PATH, "rb")
        SendData = file.read(1024)
        while SendData:
            data = {
                "action": "fullblockshash",
                "byte": (SendData.decode(encoding="iso-8859-1")),
            }
            if node is None:
                self.send(data, c_type=2)
            else:
                self.send_client(node, data, c_type=2)

            SendData = file.read(1024)

            if not SendData:
                data = {"action": "fullblockshash", "byte": "end"}
                if node is None:
                    self.send(data, c_type=2)
                else:
                    self.send_client(node, data, c_type=2)

    def send_full_blockshash_part(self, node=None):
        the_TEMP_BLOCKSHASH_PART_PATH = self.TEMP_BLOCKSHASH_PART_PATH
        file = open(the_TEMP_BLOCKSHASH_PART_PATH, "rb")
        SendData = file.read(1024)
        while SendData:
            data = {
                "action": "fullblockshash_part",
                "byte": (SendData.decode(encoding="iso-8859-1")),
            }
            if node is None:
                self.send(data, c_type=2)
            else:
                self.send_client(node, data, c_type=2)

            SendData = file.read(1024)

            if not SendData:
                data = {"action": "fullblockshash_part", "byte": "end"}
                if node is None:
                    self.send(data, c_type=2)
                else:
                    self.send_client(node, data, c_type=2)

    def get_full_chain(self, data, node, hash_of_data=""):
        self.logger.info(f"Getting full chain with {hash_of_data}")
        get_ok = False

        if not os.path.exists(self.TEMP_BLOCK_PATH):
            get_ok = True
        else:
            system = GetBlock(custom_TEMP_BLOCK_PATH=self.TEMP_BLOCK_PATH, reset=self.custom_variables)
            if node.id == system.dowload_true_block:
                get_ok = True

        if get_ok:
            if str(data["byte"]) == "end":
                move(self.LOADING_BLOCK_PATH, self.TEMP_BLOCK_PATH)

                from naruno.consensus.consensus_main import consensus_trigger
                from naruno.lib.perpetualtimer import perpetualTimer

                system = GetBlock(custom_TEMP_BLOCK_PATH=self.TEMP_BLOCK_PATH,
                                  get_normal_block=True, reset=self.custom_variables)

                ChangeTransactionFee(system)

                perpetualTimer(system.consensus_timer, consensus_trigger, the_consensus=True)
                SaveBlock(
                    system,
                    custom_TEMP_BLOCK_PATH=self.TEMP_BLOCK_PATH,
                    custom_TEMP_ACCOUNTS_PATH=self.TEMP_ACCOUNTS_PATH,
                    custom_TEMP_BLOCKSHASH_PATH=self.TEMP_BLOCKSHASH_PATH,
                    custom_TEMP_BLOCKSHASH_PART_PATH=self.
                    TEMP_BLOCKSHASH_PART_PATH,
                )

            else:
                os.chdir(get_config()["main_folder"])
                file = open(self.LOADING_BLOCK_PATH, "ab")

                file.write((data["byte"].encode(encoding="iso-8859-1")))
                file.close()

    def get_full_blockshash(self, data, node, hash_of_data=""):
        self.logger.info(f"Getting full blockshash with {hash_of_data}")
        the_TEMP_BLOCKSHASH_PATH = self.TEMP_BLOCKSHASH_PATH
        get_ok = False

        if not os.path.exists(the_TEMP_BLOCKSHASH_PATH):
            get_ok = True
        else:
            system = GetBlock(custom_TEMP_BLOCK_PATH=self.TEMP_BLOCK_PATH, reset=self.custom_variables)
            if node.id == system.dowload_true_block:
                get_ok = True

        if get_ok:
            if str(data["byte"]) == "end":
                blockshash_db.set("blockshash", None)
                with contextlib.suppress(FileNotFoundError):
                    os.remove(the_TEMP_BLOCKSHASH_PATH)
                move(self.LOADING_BLOCKSHASH_PATH, the_TEMP_BLOCKSHASH_PATH)
            else:
                os.chdir(get_config()["main_folder"])
                file = open(self.LOADING_BLOCKSHASH_PATH, "ab")
                file.write((data["byte"].encode(encoding="iso-8859-1")))
                file.close()

    def get_full_blockshash_part(self, data, node, hash_of_data=""):
        self.logger.info(f"Getting full blockshash part with {hash_of_data}")
        the_TEMP_BLOCKSHASH_PART_PATH = self.TEMP_BLOCKSHASH_PART_PATH
        get_ok = False

        if not os.path.exists(the_TEMP_BLOCKSHASH_PART_PATH):
            get_ok = True
        else:
            system = GetBlock(custom_TEMP_BLOCK_PATH=self.TEMP_BLOCK_PATH, reset=self.custom_variables)
            if node.id == system.dowload_true_block:
                get_ok = True

        if get_ok:
            if str(data["byte"]) == "end":
                blockshash_db.set("blockshash_part", None)
                with contextlib.suppress(FileNotFoundError):
                    os.remove(the_TEMP_BLOCKSHASH_PART_PATH)
                move(self.LOADING_BLOCKSHASH_PART_PATH,
                     the_TEMP_BLOCKSHASH_PART_PATH)
            else:
                os.chdir(get_config()["main_folder"])
                file = open(self.LOADING_BLOCKSHASH_PART_PATH, "ab")
                file.write((data["byte"].encode(encoding="iso-8859-1")))
                file.close()

    def get_full_accounts(self, data, node, hash_of_data=""):
        self.logger.info(f"Getting full accounts with {hash_of_data}")
        the_TEMP_ACCOUNTS_PATH = self.TEMP_ACCOUNTS_PATH
        the_LOADING_ACCOUNTS_PATH = self.LOADING_ACCOUNTS_PATH
        get_ok = False

        if not os.path.exists(the_TEMP_ACCOUNTS_PATH):
            get_ok = True
        else:
            system = GetBlock(custom_TEMP_BLOCK_PATH=self.TEMP_BLOCK_PATH, reset=self.custom_variables)
            if node.id == system.dowload_true_block:
                get_ok = True

        if get_ok:
            if str(data["byte"]) == "end":
                accounts_db.set("accounts", None)
                with contextlib.suppress(FileNotFoundError):
                    os.remove(the_TEMP_ACCOUNTS_PATH)
                move(the_LOADING_ACCOUNTS_PATH, the_TEMP_ACCOUNTS_PATH)
            else:
                os.chdir(get_config()["main_folder"])
                file = open(the_LOADING_ACCOUNTS_PATH, "ab")
                file.write((data["byte"].encode(encoding="iso-8859-1")))
                file.close()

    @staticmethod
    def send_transaction(
        tx,
        custom_current_time=None,
        custom_sequence_number=None,
        custom_balance=None,
        except_client=None,
        custom_server=None,
    ):
        """
        Sends the given transaction to UNL nodes.
        """
        if not tx.signature == "NARUNO":
            data = {
                "action": "transactionrequest",
                "sequence_number": tx.sequence_number,
                "txsignature": tx.signature,
                "fromUser": tx.fromUser,
                "to_user": tx.toUser,
                "data": tx.data,
                "amount": tx.amount,
                "transaction_fee": tx.transaction_fee,
                "transaction_time": tx.transaction_time,
                "custom_current_time": custom_current_time,
                "custom_sequence_number": custom_sequence_number,
                "custom_balance": custom_balance,
            }
            the_server = server.Server if custom_server is None else custom_server
            the_server.send(data, except_client=except_client, c_type=1)

    def get_transaction(self, data, node, hash_of_data=""):
        self.logger.info(f"Getting transaction with {hash_of_data}")
        if self not in naruno.node.server.server.tx_signature_list:
            naruno.node.server.server.tx_signature_list[self] = []

        if data["txsignature"] in naruno.node.server.server.tx_signature_list[self]:
            self.logger.debug("Transaction is already getted")
            return

        naruno.node.server.server.tx_signature_list[self].append(data["txsignature"])
        
        the_transaction = Transaction(
            data["sequence_number"],
            data["txsignature"],
            data["fromUser"],
            data["to_user"],
            data["data"],
            data["amount"],
            data["transaction_fee"],
            data["transaction_time"],
        )
        block = GetBlock(custom_TEMP_BLOCK_PATH=self.TEMP_BLOCK_PATH, reset=self.custom_variables)
        custom_current_time = None
        custom_sequence_number = None
        custom_balance = None
        if self.custom_variables:
            custom_current_time = data["custom_current_time"]
            custom_sequence_number = data["custom_sequence_number"]
            custom_balance = data["custom_balance"]

        GetTransaction(
                block,
                the_transaction,
                custom_current_time=custom_current_time,
                custom_sequence_number=custom_sequence_number,
                custom_balance=custom_balance,
                custom_PENDING_TRANSACTIONS_PATH=self.
                PENDING_TRANSACTIONS_PATH,
                except_client=node,
                custom_server=self)
        
    def send_block_to_other_nodes(self,
                                  node=None,
                                  sync=False,
                                  hash_of_data=""):
        """
        Sends the block to the other nodes.
        """
        self.logger.info(f"Sending block to other nodes with {hash_of_data}")
        if node is None or sync:
            self.logger.info("Sync process started")
            self.send_full_accounts(node=node)
            self.send_full_blockshash(node=node)
            self.send_full_blockshash_part(node=node)
            self.send_full_chain(node=node)
        else:
            self.logger.info("Node appended to sync_clients")
            self.sync_clients.append(node)

    def get_ip(self):
        """
        Returns the IP address of the socket in this class.
        """
        ip = self.sock.getsockname()[0]
        return ip



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import copy
import json
import os
from hashlib import sha256
import contextlib

from naruno.lib.config_system import get_config
from naruno.lib.kot import KOT

mytransactions_db = KOT("mytransactions",
                        folder=get_config()["main_folder"] + "/db")

import naruno


def SaveMyTransaction(transaction_list, clear=False):
    """
    Saves the transaction_list to the transaction db.
    """
    if type(transaction_list) is list:
        entry_name_list = []
        for tx in transaction_list:
            name = copy.copy(tx[0].signature.encode("utf-8"))
            if tx[0].signature == b"":
                name = "empty".encode("utf-8")

            entry_name_list.append(sha256(name).hexdigest())

        new_dict = {
            tx[0].signature: {
                "tx": tx[0].dump_json(),
                "validated": tx[1],
                "sended": tx[2],
            }
            for tx in transaction_list
        }

        if clear:
            record = mytransactions_db.get_all() if naruno.transactions.my_transactions.get_my_transaction.mytransactions_db_ram == {} else naruno.transactions.my_transactions.get_my_transaction.mytransactions_db_ram
            for entry in copy.copy(record):
                if not entry.endswith("validated") and not entry.endswith(
                        "sended"):
                    if entry not in str(entry_name_list):
                        mytransactions_db.delete(entry)
                        mytransactions_db.delete(entry + "validated")
                        mytransactions_db.delete(entry + "sended")
                        with contextlib.suppress(KeyError):
                            naruno.transactions.my_transactions.get_my_transaction.mytransactions_db_ram.pop(entry)
                        with contextlib.suppress(KeyError):
                            naruno.transactions.my_transactions.get_my_transaction.mytransactions_db_ram.pop(entry + "validated")
                        with contextlib.suppress(KeyError):
                            naruno.transactions.my_transactions.get_my_transaction.mytransactions_db_ram.pop(entry + "sended")

        transaction_list = new_dict

        for tx in transaction_list:
            name = copy.copy(tx.encode("utf-8"))
            if tx == b"":
                name = "empty".encode("utf-8")
            hash_of_name = sha256(name).hexdigest()
            mytransactions_db.set(
                hash_of_name, transaction_list[tx]["tx"])
            naruno.transactions.my_transactions.get_my_transaction.mytransactions_db_ram[hash_of_name] = transaction_list[tx]["tx"]
            

            if transaction_list[tx]["validated"]:
                mytransactions_db.set(
                    hash_of_name + "validated", True)
                naruno.transactions.my_transactions.get_my_transaction.mytransactions_db_ram[hash_of_name+ "validated"] = True
            if transaction_list[tx]["sended"]:
                mytransactions_db.set(
                    hash_of_name + "sended", True)
                naruno.transactions.my_transactions.get_my_transaction.mytransactions_db_ram[hash_of_name+ "sended"] = True

    elif type(transaction_list) is dict and transaction_list != {}:
        hash_of_name = sha256(transaction_list[0].signature).hexdigest()
        mytransactions_db.set(
            hash_of_name,
            transaction_list[0].dump_json(),
        )
        naruno.transactions.my_transactions.get_my_transaction.mytransactions_db_ram[hash_of_name] = transaction_list[0].dump_json()

        if transaction_list[1]:
            mytransactions_db.set(
                hash_of_name + "validated", True)
            naruno.transactions.my_transactions.get_my_transaction.mytransactions_db_ram[hash_of_name+ "validated"] = True
        if transaction_list[2]:
            mytransactions_db.set(
                hash_of_name+ "sended",True)
            naruno.transactions.my_transactions.get_my_transaction.mytransactions_db_ram[hash_of_name+ "sended"] = True



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import contextlib
import json
import os
import time

from naruno.accounts.account import Account
from naruno.accounts.save_accounts import SaveAccounts
from naruno.blockchain.block.block_main import Block
from naruno.blockchain.block.blocks_hash import SaveBlockshash
from naruno.blockchain.block.blocks_hash import SaveBlockshash_part
from naruno.config import TEMP_BLOCK_PATH
from naruno.consensus.rounds.round_1.process.transactions.checks.duplicated import \
    Remove_Duplicates
from naruno.lib.config_system import get_config
from naruno.lib.kot import KOT
from naruno.lib.log import get_logger
from naruno.lib.settings_system import the_settings
from naruno.transactions.cleaner import Cleaner
from naruno.transactions.pending.get_pending import GetPending
import naruno

block_db = KOT("block_db", folder=get_config()["main_folder"] + "/db")

logger = get_logger("BLOCKCHAIN")



serial_testers = [
    "2844854d58334e66540097479439be65e0aa4efc",
    "7f53b1bb58032bc0212b8f1abc9fbb6510a16671",
    "8e0a8f580c43dd2b33d62afbfa514b3cbe262061",
    "0022b1a2c50249be6e760be4f3c2c0d9ed467fc3",
    "65abccda166c7f730376a0c15753942392b21d73",
    "c8ecc03e12e9b1975851ac4bac2159e2210c8402",
    "80df451fc2653ee4bb85230f2e9800f07527f778",
    "f51b88c59db9cac5e933ac8827bd7f10dc543ae3",
    "59bd07543a4aaf2e45a9b8b8f29e372de79d47ce",
    "7c697b908b77064d70ef4526a658eb9e2e5a9ee6",
    "5764cfd11addb37353c7e159a909e90f1de3c5e9",
    "e4ec7b50fe90c117ad37b41b01e5059c1c903ce0",
    "fd29f1d5b573796d4bd614c3c52738bbfcc5e340",
    "b01a9cdd81bad9cefc43733b5146c2a1b1b4dd6c",
    "c8dca04b72bc69b4e3ac3067f6d0174f2dcfe680",
    "755089190444e4af31aa4641c8d6e03254014b11",
    "68b777ebba65e8ab4af3d6bdb0837abc3ae177d9",
    "f641261d1527cade37371957afe04eb2d7e6b471",
    "7c7bf0510137fc29f09f2df2e38a45146567df7d",
    "1aa29b64943c955e2c1dfa5b734a5c617cdd4a8d",
    "d865a195c73d06313d72dcf3579d9920e4b4584b",
    "60d4357a85a2113467860455e5ccdb7b127a97dd",
    "bd679bd9bf83a00766cb0ae29ab0f9aae7f2d73e",
    "2b0d9405d9f8a6c658608dd2255528e0a12f4f16",
    "884aaa97c4121bbee5a3d79a312900c090bbbff5",
    "29b7ab3aeafedef1c3ad81323473ade7877ad698",
    "cc9ac91de747f657c1a2b6c9e071cf53833e357d",
    "52185af45f343636dcb2e88843a12e24aad3be9c",
    "73b02b4a4715c96eec46823b5fb1e1f401dcd1cb",
    "4b59062c6d3840a023fdc8aaa5c0a87f05aab62e",
    "8ed1aea1b9ebe2e4cfc9b9b03b53ac892bc63119",
    "6c9189efacd2c9b2bdbde200b5bae8cb022a9bd7",
    "ddf8a128b477f00167367f8d7e7816412c81feae",
    "58f9bda368fb61dcdc177af0daea1eb46b53ed5",
    "16bdffb8012f79c7818ee67054994fe128012cc4",
    "eae2c4f8ff12c10e0a3b0c8a5a677a2b9e50309e",
    "fcd23380b3ef500b205424c435837b8367e7e0a6",
    "1e65a85a0716c53fca521d7cbf9096d2f9a18ab4",
    "c010bb210da143a01d0d2ae914fd033c549e33c7",
    "1c8d4b723b67165847e3b11e3029748e6ebf5139",
    "b2345bda4ff22d0c51f27c4d470c63f63730ac77",
    "17008875dc5a6c9102a5c2aa1e03249d7f576237",
    "71fa49f318437ac5f3a72f6048f0f15dff974cce",
    "095fef5e67fad85a0cd40517d46316741f7b7e5c",
    "b1956275dadc0e91952b7991d692457c8699c73b",
    "d5a262b8719540b95e742838abf50a67c3972d82",
    "e493b8b92c3271af170064bcf42f706272bce631",
    "51a5563ecd11b49c19f2f11d3bb8ac1a0f75e4ea",
    "1243e02aa77ca080edd879b031fee5d2858ef248",
    "069f4682619925d4862c9597ac501c67b39e0608",
    "bdf78ffeed1cb2e37e2cd1e444ee6cf552b62e40",
    "6b31a60a3a73aa7b80b5a63fed63fcfdb85ee863",
    "a426fad6820c77d5d0c730c7f13bd1d668e17b45",
    "90f248e2d1143fd3995ebc2de856115d305ee46c",
    "6c3b4fd7a8f619194f13bacfb83bbeee91bf1ce0",
    "746c2ed9eae6210533854c217586b27330f80417",
    "329506e24ef995ec3e2759242ce52033a41be14a",
    "a02a508328575e913a09a9dd28a11145e62c9865",
    "9abb7a1219a20d6e55f8e4d5ae6ec0e89c2b4770",
    "1bb50e042ce653ec8501f47d712dc1ded0c6374a",
    "d3cbae4148f79953f5b86c8ab5c70282cbf398b7",




]
individuals = [
    "df78cbab1f97ec8115ad2d2a48f7e5cbbe754bb8",
    "55de207a538855b4da2d60325e8afadc3b3caa04",
    "86e9a2454e2d2bd12f56ef37e39d026608013e72",
    "2f58be5d152490affa05a7b0fd3cef8c195dae6d",
    "a26536e07f3c2a850fb2b63cbe99d84589674634",
    "0be5c9cd8bf68cafeec3a2d5d51678923780d3ff",
    "82d87a6bfd279d30ad4894912eae2efacd4d46d6",
    "f6e4955a8077ae5ed7d95014b41f22dcee6c0d76",
    "73672aafc1890fc18d9b88105380b396eca799a5",
    "83a15e056f98305418ee9ea26caf664c3d020040",
    "b1df8deda30d4f88cb905ecd57ed0fc7f2021d00",
    "ec29c2e01987796a3677da2e3a9b4a098b93b89a",
    "887af3d44bfe39005b4cc480c2b03a11c2fb8b63",
    "17d3d3e20bd84ddf6e3ed85fa693c12654f174eb",
    "1da75d769ab3604abc04763d20dc3f70bf1c69b8",
    "d7eee170a14b99e37a3e3fc6d375d1d28bddf62b",
    "d7f20b7990cc593e248f1d0dd31dd7a235a40d9a",
    "75f7e0e090834c959f6538b1a4c80f03be410bdd",
    "d37cb2c0df30965f2bc12cea040386e70e32402b",
    "0af54b3f47fc1577688e7c2c227672f610cad292",
    "96bb7bac1af450ea0c17300ea4f61b1cd0b88b6d",
    "709bf017a48d2f02bb5d5d8d205ccf39d8205b4a",
    "5b943d77a8b7e66aa60b98a2197f2197db4f464b",
    "bba1e2f5871c02b130416229c02dd54fc404cd21",
    "dcf715a42784bfedf009b515ada46c1946a3339b",
    "0af54b3f47fc1577688e7c2c227672f610cad292",
    "0af54b3f47fc1577688e7c2c227672f610cad292",
    "0af54b3f47fc1577688e7c2c227672f610cad292",
    "cbdeeab5577f6f8693e571494e61dc0f356d9d09",
    "09375da881e7d546b8520a7e39160fc05eb60ce1",
    "b3232a5ed1d57c339fb3a258be40c51e221b48af",
    "7be509d428fc848485683a599cc4bbd958bd6df3",
    "2b76e7031353e9cbf62e2d259b54b98673f997c9",
    "614deef9abde68a0dab39fe5ee9f9cb2f9e8773a",
    "44281e3b85497d41d8c99f3b8b66a3e88e354ca1",
    "5db89649f397be7d3787cfca2e9f39245ef5f8e1",
    "b3e6e03b6c3f6ea8989177d23a9a055e2a05659b",
    "bc512bab30167fe84c84208be8759bf9839d7815",
    "138c27b39d5789319aa9a0c00262cb5360d146d0",
    "0dbe3d640ee1632b222154197a23319870f3d12d",
    "4a1b3198530952ee851722b105770ef21c59e472",
    "541cb5cc142e94aed187f83ee0d91a1470db3023",
    "caefad8f89b701d8087945780ad24fe0f993ae2f",
    "8f571575a492ef27f0fe02916303ea867f50876d",
    "508c9565bac49d318eb5c67f6e6e0354acbc4edf",
    "c7b71bd5100d2081de69497776064589e98c4fe1",
    "c58b9d12474108867f4a0924f4c521119a57d62f",
    "0d435e7fb86a46f4849a4eb41f683ab322f03d6b",
    "2E6A70110890b7A46F7857E2866023043CeF4680",
    "078e6dd15b33411462d284baf4c901cfd828558c",
    "88b73c67bf3aab4764d627cc9aa1150c6ae97794",
    "450305d3bb1a0fd7343824f136177220d436c046",
    "e7b72ca5af68bc5b8ddd98e4187c68efcfbec5cd",
    "b356e5edd49015dad8b69c607130588512380f24",
    "6af5cba8cf348e92717ba1e08132e19220d2239d",
    "78b830a53b4169f815da314d7dec8bbd935eab1e",
    "d2bbee821194b41733a9aac650274690c11923c6",
    "21d81b6bf3e4a6f693dfbcbeff5ba0af69130939",
    "88f4a4ace0afb7621c6e33042bdc3674c37a5b87",
    "f2e63a212a1d318f428a67eaac537a33bf9fa3ee",
    "3e9ec4425b3d40bb0fbe630ce6353922fc58d174",
    "899de6fc19d9e34beab208753fffb71fb4e4d52c",
    "f2e63a212a1d318f428a67eaac537a33bf9fa3ee",
    "c3dec99a1e27d739e25102e528a673dc9efd3217",
    "d851fbd6e38c4e9f91be7768586765ce6cf1e3ac",
    "37f7bfc9c0b7cf67a8cb8e3e757f3be00505df6f",
    "ca5f6b56a4ddda561618b4a1460040ba2089d6d0",
    "a517e581e265b6690ef1d58331f685df9276e7e4",
    "f9ac49630d4e8319cb860e1492154585cc7603ce",
    "1353f11e29fecd0f023ffbad20e13fbc8dbcc417",
    "2cc4abe152c4b6a2a5108e76e2895656e82c33dc",
    "c42e8c1d91de134c9e868877f495bbe2a2f0a2c3",
    "9872f526e450a21782cc9db8b18867dc5a26b93a",
    "73a4a4a4905bb4352633c49ec8641f0e6555c888",
    "33e5a77ccef17133d0a87522cafca3c6cf9d4512",
    "79948de58a4265e58199489176730ebe121c6faa",
    "490fcf6651419d2a14d6f7ada52ac2bae51f9247",
    "d2761b3f2a80e7c854a31088f5a4c48e0c30f630",
    "198796aeb4b896d43a6b1bc82597349de6a231cc",
    "078166eb3a997091b65f4fb25f9e434b9d13b7f3",
    "97d46be7a5e281e7bb71ac8ef1ee0eeca716a8af",
    "0e4004f212eb9b38443cdafa37d8a41aaa06da46",
    "29dd1a158ab266ebae6f75e1a09c676f2fef0726",
    "b5274cc8e238f855687d15efd1863d47aed38fdd",
    "948812de96640a783dcb2696c9e04570382f24c3",
    "1e43c76cdded5255f95a0ee0f886133c5cb953cf",
    "813a3637cf67613b4755c0e849ea05b760f83c3a",
    "2d9ac3d72feeafa2f86b4ffb57d4c8c33993d57",
    "49b9529a5eb35188b9bcc938e8a9bc23052aa4cb",
    "df7c8bd0bcee84959747547199f8f64af3c71af7",
    "aaefe424bc7cf62c9ea20a8a111c56fdb7efa05b",
    "ae431c1971be9d605c150f39454cbcdce0e7ec07",
    "fe44fb36a82bc177794108d37e97b696ba982257",
    "2b29e43956530c93be2c90022dfcc0dfff0c6419",
    "7e8e4798ef5607834522712cf6ebf7f6625a2eed",
    "fc94dc3d2ee57aea37dee8c6cc66fdef60bef519",
    "d575221dffadf3d626c53d5e2caea06653d0c5b7",
    "683b99fb0fc3a5c0e70f97ce16c8c803bb3cf828",
    "99b9a4552a5484bce18d01dd6e165d497099b754",
    "2ae08ace0c213cd17d08e9a97798137abaa4bf13",
    "9039cc13c0fc89ee097fe9f608cf387da3971c8f",
    "28e368a5cc6b4f3b7f09da5cf16bef1dc25c97b8",
    "dc0fcf7851f565369c0c616e704ce0af6ecc61db",
    "0944281ea3c0e5290d2d56547aec447fb823e3f2",
    "82a347178b62a3203aef15cbad2e6c70a319ba46",
    "c35a9acc440a980b136c5cf0b8a801b03a97e2f1",
    "alb78e89d508ea358d32db79c20e444175551d7e",
    "49b9529a5eb35188b9bcc938e8a9bc23052aa4cb",
    "6256d81a0d280048c6907f20c2654b0ca027863c",
    "c6618b1c64fac65fe58f8da8fc75f499b744f55e",
    "cd02e3d2208b556d1115a2b8f62d2cc27af54389",
    "ffc5219f0108bf02654f949f18037d68264c42bc",
    "5ab039bdb7c6dfe57507834874f7494165a1f12b",
    "af32b70e92506ed2128e0f58f233129d24dc1723",
    "22974433614765cdc42fc94108dc0a0d5349bc73",
    "2b0c8c2bc1c725056696c896425e8264fae2760f",
    "0f5902f6285a9b7723c565d756e89f7a0ba9a3b3",
    "a94976beb574e53bdc1e86a36f6415067123de86",
    "9a46fd0899fcaf5fef9e7cd277b1bf1864827d24",
    "cb9314ad3ecd2cff3f7aa1a1fc04f13c5b026270",
    "f5647d2072bdfa89b6bd05e7051bf9e57bac5dd2",
    "8029f657fe71de6c42491c9969f795b4be68fee7",
    "05c19214daba9aacab274f0c89a93721919d80a1",




]

specials = ["40a175793d9b7082157fd3e1632fc25c9f3f5234"]

special_testers = [
    "d78101e2cb261efab28ee54a0e6d716820a1bab1",
    "40a175793d9b7082157fd3e1632fc25c9f3f5234",
    "9febe30ed21991fdca2e84d1a42d594c1a02ab36",
    "f687088eff5423f913f03b677f97a9c405f7549a",
    "c923c646f2d73fcb8f626afacb1a0ade8d98954a"
    ]


def SaveBlock(
    block: Block,
    custom_TEMP_BLOCK_PATH=None,
    custom_TEMP_ACCOUNTS_PATH=None,
    custom_TEMP_BLOCKSHASH_PATH=None,
    custom_TEMP_BLOCKSHASH_PART_PATH=None,
    delete_old_validating_list=False,
    just_save_normal=False,
    dont_clean=False,
):
    """
    Saves the current block to the TEMP_BLOCK_PATH.
    """
    if not dont_clean:
        if not block.round_1 and not block.round_2:
            cleaned = Cleaner(block, pending_list_txs=GetPending())
            block.validating_list = cleaned[0]

            block = Remove_Duplicates(block)
            block.validating_list = sorted(block.validating_list,
                                           key=lambda x: x.fromUser)

    logger.info("Saving block to disk")
    logger.debug(
        f"Block#{block.sequence_number + block.empty_block_number}:{block.empty_block_number}: {block.dump_json()}"
    )

    if block.first_time:
        accounts_list = [Account(block.creator, block.coin_amount)]
        baklava_test_net_users = []

        baklava_test_net_users.extend([Account(i,(2 * block.minumum_transfer_amount) + block.transaction_fee * 100,) for i in serial_testers])
        baklava_test_net_users.extend([Account(i,(2 * block.minumum_transfer_amount) + block.transaction_fee * 100,) for i in special_testers])
        baklava_test_net_users.extend([Account(i,(2 * block.minumum_transfer_amount) + block.transaction_fee * 100,) for i in individuals])
        baklava_test_net_users.extend([Account(i,(2000 * block.minumum_transfer_amount) + block.transaction_fee * 100,) for i in specials])



        if the_settings()["baklava_users"]:
            accounts_list.extend(baklava_test_net_users)
        SaveAccounts(
            accounts_list,
            custom_TEMP_ACCOUNTS_PATH=custom_TEMP_ACCOUNTS_PATH,
        )
        SaveBlockshash(
            block.previous_hash,
            custom_TEMP_BLOCKSHASH_PATH=custom_TEMP_BLOCKSHASH_PATH,
        )
        SaveBlockshash_part(
            [block.previous_hash],
            custom_TEMP_BLOCKSHASH_PART_PATH=custom_TEMP_BLOCKSHASH_PART_PATH,
        )
        block.first_time = False
    the_TEMP_BLOCK_PATH = (TEMP_BLOCK_PATH if custom_TEMP_BLOCK_PATH is None
                           else custom_TEMP_BLOCK_PATH)
    secondly_situation = 0
    if block.round_1:
        secondly_situation += 1
    if block.round_2:
        secondly_situation += 1
    highest_the_TEMP_BLOCK_PATH = (
        the_TEMP_BLOCK_PATH + "-" +
        str(block.sequence_number + block.empty_block_number) + "-" +
        str(len(block.validating_list)) + "-" + str(secondly_situation) + "-" +
        str(time.time()))
    logger.info(f"Saving block to {highest_the_TEMP_BLOCK_PATH}")

    if delete_old_validating_list:
        os.chdir(get_config()["main_folder"])
        for file in os.listdir("db/"):
            if ("db/" + file).startswith(the_TEMP_BLOCK_PATH) and not (
                    "db/" + file) == the_TEMP_BLOCK_PATH:
                with contextlib.suppress(IndexError):
                    number = int((("db/" + file).replace(the_TEMP_BLOCK_PATH,
                                                        "")).split("-")[1])
                    high_number = int(
                        (("db/" + file).replace(the_TEMP_BLOCK_PATH,
                                                "")).split("-")[2])
                    secondly_situation_number = int(
                        (("db/" + file).replace(the_TEMP_BLOCK_PATH,
                                                "")).split("-")[3])
                    if (number == block.sequence_number + block.empty_block_number
                            and high_number != len(block.validating_list)
                            and secondly_situation_number == 1):
                        with contextlib.suppress(FileNotFoundError):
                            logger.info(f"Deleting old validating list: {file}")
                            block_db.delete("db/"+file)

    for file in os.listdir("db/"):
        if ("db/" + file).startswith(the_TEMP_BLOCK_PATH) and not (
                "db/" + file) == the_TEMP_BLOCK_PATH:
            with contextlib.suppress(IndexError):
                number = int((("db/" + file).replace(the_TEMP_BLOCK_PATH,
                                                     "")).split("-")[1])  # seq
                high_number = int(
                    (("db/" + file).replace(the_TEMP_BLOCK_PATH,
                                            "")).split("-")[2])  # val
                if number < block.sequence_number + block.empty_block_number:
                    with contextlib.suppress(FileNotFoundError):
                        logger.info("Removing " + "db/" + file)
                        block_db.delete("db/"+file)

    block_db_path_first = os.path.join(get_config()["main_folder"],
                                       the_TEMP_BLOCK_PATH)

    naruno.blockchain.block.get_block.the_ram_block[the_TEMP_BLOCK_PATH] = block
    block_db.set(the_TEMP_BLOCK_PATH,
                 block.dump_json(),
                 custom_key_location=block_db_path_first)

    if not just_save_normal:
        block_db_path_second = os.path.join(get_config()["main_folder"],
                                            highest_the_TEMP_BLOCK_PATH)
        block_db.set(highest_the_TEMP_BLOCK_PATH,
                     block.dump_json(),
                     custom_key_location=block_db_path_second)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from naruno.blockchain.block.block_main import Block
from naruno.lib.log import get_logger

logger = get_logger("CONSENSUS_FIRST_ROUND")


def find_newly(block: Block, temp_validating_list: list) -> list:
    """
    Finds not validated new transaction in our block remove than and return as a new list
    """
    logger.info("Find new transactions process is started")
    newly_added_list = []
    logger.debug(f"First newly_added_list: {newly_added_list}")

    for my_validating_list in block.validating_list[:]:
        ok = any(
            (my_validating_list.signature == my_temp_validating_list.signature)
            for my_temp_validating_list in temp_validating_list[:])

        block.validating_list.remove(my_validating_list)
        logger.info(
            f"tx: {my_validating_list} will removed fron block.validating_list"
        )
        if not ok:
            logger.info(
                f"tx: {temp_validating_list} will added to temp validating list"
            )
            newly_added_list.append(my_validating_list)
    logger.debug(f"End newly_added_list: {newly_added_list}")
    return newly_added_list



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import json
import os

from naruno.config import PENDING_TRANSACTIONS_PATH
from naruno.lib.config_system import get_config

from naruno.transactions.transaction import Transaction

from naruno.transactions.pending.save_pending import pendingtransactions_db
from naruno.lib.kot import KOT

def GetPending(custom_PENDING_TRANSACTIONS_PATH=None):
    if custom_PENDING_TRANSACTIONS_PATH == PENDING_TRANSACTIONS_PATH:
        custom_PENDING_TRANSACTIONS_PATH = None
    the_pending_list = []
    all_records = pendingtransactions_db.get_all() if custom_PENDING_TRANSACTIONS_PATH is None else KOT("pendingtransactions"+custom_PENDING_TRANSACTIONS_PATH, folder=get_config()["main_folder"] + "/db").get_all()
    for entry in all_records:
        the_pending_list.append(Transaction.load_json(all_records[entry]))
    return sorted(the_pending_list, key=lambda x: x.signature)


def GetPendingLen(custom_PENDING_TRANSACTIONS_PATH=None):


    return pendingtransactions_db.get_count() if custom_PENDING_TRANSACTIONS_PATH is None else KOT("pendingtransactions"+custom_PENDING_TRANSACTIONS_PATH, folder=get_config()["main_folder"] + "/db").get_count()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import contextlib
import json
import socket
import time
from hashlib import md5
from threading import Thread
import traceback

from naruno.blockchain.block.block_main import Block
from naruno.lib.log import get_logger



class client(Thread):

    def __init__(self, socket, address, node_id, server, test=False,c_type=0):
        Thread.__init__(self)
        self.server = server
        self.socket = socket
        self.host = address[0]
        self.port = address[1]
        self.id = node_id
        self.candidate_block = None
        self.candidate_block_hash = None
        self.candidate_block_history = []
        self.candidate_block_hash_history = []

        self.logger = get_logger(
            f"NODE_{self.server.host}_{self.server.port}_SOCK_{self.host}_{self.port}"
        )


        a_block = Block("onur")
        self.logger.debug(f"Block max_data_size: {a_block.max_data_size}")
        self.logger.debug(f"Block max_tx_number: {a_block.max_tx_number}")        
        buffer_size = 6525 + int(
            (a_block.max_data_size // a_block.max_tx_number) * 1.5)
        buffer_size_2 = 6525 + int(
            (a_block.max_data_size // a_block.max_tx_number) * 1.5)
        buffer_size_3 = 6525 + int(
            (a_block.max_data_size // a_block.max_tx_number) * 5)  
        buffer_size_4 = 6525 + int(
            (a_block.max_data_size // a_block.max_tx_number) * 0.5)  

        self.c_type = c_type
        if self.c_type == 0:
            self.buffer_size = buffer_size
        elif self.c_type == 1:
            self.buffer_size = buffer_size_2          
        elif self.c_type == 2:
            self.buffer_size = buffer_size_3   
        elif self.c_type == 3:
            self.buffer_size = buffer_size_4


        self.logger.info(f"Connection established with {self.id}")



        self.logger.debug("Test mode: " + str(test))
        self.logger.debug("Buffer size: " + str(self.buffer_size))

        self.running = True
        if not test:
            self.start()


    def threaded_receive(self, data):
                data = data.decode("utf-8")
                with contextlib.suppress(json.decoder.JSONDecodeError):
                    data = json.loads(data)
                with contextlib.suppress(Exception):
                    del data["buffer"]
                try:
                    hash_of_data = md5(str(data).encode()).hexdigest()[:6]
                    self.logger.debug(
                        f"Received data ({hash_of_data}): {data}")
                    self.server.get_message(self, data, hash_of_data)
                except Exception as e:
                    traceback.print_exc()
                    self.logger.error(f"Error while processing data: {e}")
                    self.logger.error(f"Data: {data}")

    def run(self):
        self.socket.settimeout(10.0)
        while self.running:
            with contextlib.suppress(socket.timeout):
                data = self.socket.recv(self.buffer_size)

                if not data:
                    break

                Thread(target=self.threaded_receive, args=(data,)).start()

            time.sleep(0.01)

    def stop(self):
        self.running = False
        self.socket.close()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import contextlib
import copy
import threading


from naruno.blockchain.block.block_main import Block
from naruno.lib.log import get_logger
from naruno.node.server.server import server
from naruno.transactions.pending.delete_pending import DeletePending
from naruno.transactions.pending.get_pending import GetPending
from naruno.transactions.pending.save_pending import SavePending

logger = get_logger("TRANSACTIONS")




def PendingtoValidating(block: Block):
    """
    Adds transactions to the verification list
    if there are suitable conditions.
    """
    logger.info("Pending to validating transfer process is started")
    first_validating_list_len = len(block.validating_list)
    first_max_tx_number = block.max_tx_number
    logger.debug(f"Currently tx amount: {first_validating_list_len}")
    logger.debug(f"Validating list capacity: {first_max_tx_number}")

    pending_list_txs = GetPending()
    logger.debug(f"Pending list is got: {pending_list_txs}")




    first_situation = copy.copy(block.validating_list)
    the_list_of_tx = pending_list_txs + first_situation
    
    block.validating_list = []

    logger.debug(f"Pending list is sent to server")
    if len(block.validating_list) < block.max_tx_number:
        logger.debug("List is not full")
        for tx in OrderbyFee(the_list_of_tx):
            logger.debug(f"TX {tx.signature} is checking")
            if len(block.validating_list) < block.max_tx_number:
                logger.info(f"tx {tx.signature} is moved to validating list")

                block.validating_list.append(tx)
                the_list_of_tx.remove(tx)
                if not tx in first_situation:
                    DeletePending(tx)

            else:
                logger.info(
                    f"TX {tx.signature} is can not moved to validating list")
    else:
        logger.info("List is full")
    
    for i in the_list_of_tx:
        if i in first_situation:
            SavePending(i)
        

    


def OrderbyFee(transactions: list):
    """
    Sorts transactions by fee.
    """

    the_transactions = copy.copy(transactions)
    the_transactions.sort(key=lambda x: x.signature, reverse=True)
    the_transactions.sort(key=lambda x: x.transaction_fee, reverse=True)
    return the_transactions



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import sqlite3
from decimal import Decimal
from naruno.accounts.account import Account
from naruno.accounts.save_accounts import SaveAccounts
from naruno.blockchain.block.shares import shares
from naruno.config import TEMP_ACCOUNTS_PATH
from naruno.consensus.rounds.round_1.process.transactions.checks.duplicated import \
    Remove_Duplicates
from naruno.lib.log import get_logger
from naruno.wallet.wallet_import import Address

logger = get_logger("TRANSACTIONS")


def ProccesstheTransaction(
    block,
    the_account_list,
    custom_TEMP_ACCOUNTS_PATH=None,
    custom_shares=None,
    custom_fee_address=None,
    dont_clean=False,
):
    """
    It performs the transactions in the block.vali list and
    puts the transactions in order.

    Queuing is required so that all nodes have the same transaction hash.
    """

    logger.info("Transaction processing started.")

    if not dont_clean:
        block = Remove_Duplicates(block)

    the_TEMP_ACCOUNTS_PATH = (TEMP_ACCOUNTS_PATH if custom_TEMP_ACCOUNTS_PATH
                              is None else custom_TEMP_ACCOUNTS_PATH)

    edited_accounts = []

    clean_list = []
    for unclear in block.validating_list:
        if not "NARUNO" in unclear.signature:
            clean_list.append(unclear)
    block.validating_list = clean_list

    the_shares = shares(
        block,
        custom_shares=custom_shares,
        custom_fee_address=custom_fee_address,
        dont_clean=dont_clean,
    )
    block.validating_list = block.validating_list + the_shares

    new_added_accounts_list = []
    account_list = []

    block.validating_list = sorted(block.validating_list,
                                   key=lambda x: x.fromUser)

    temp_validating_list = block.validating_list

    actions = []

    for trans in block.validating_list:
        logger.info(f"Transaction: {trans.__dict__}")

        touser_inlist = True
        to_user_in_new_list = False

        address_of_fromUser = Address(trans.fromUser)
        logger.debug(f"FromUser address: {address_of_fromUser}")
        the_record_account_list = []

        the_record_account_list.append([
            address_of_fromUser,
            the_account_list[address_of_fromUser][0],
            the_account_list[address_of_fromUser][1],
        ]) if address_of_fromUser in the_account_list else None
        the_record_account_list.append([
            trans.toUser,
            the_account_list[trans.toUser][0],
            the_account_list[trans.toUser][1],
        ]) if trans.toUser in the_account_list else None

        for the_pulled_account in the_record_account_list:
            account_list.append(
                Account(the_pulled_account[0], the_pulled_account[2],
                        the_pulled_account[1]))

        for Accounts in account_list:
            touser_inlist = False

            if Accounts.Address == address_of_fromUser:
                logger.debug(f"FromUser found: {Accounts.Address}")
                actions.append([
                    Accounts.Address,
                    "balance",
                    -(float(trans.amount) + trans.transaction_fee),
                ])
                actions.append([Accounts.Address, "sequence_number", 1])

            if Accounts.Address == trans.toUser:
                logger.debug(f"ToUser found: {Accounts.Address}")
                actions.append(
                    [Accounts.Address, "balance",
                     float(trans.amount)])
                touser_inlist = True

        for i in new_added_accounts_list:
            if i.Address == trans.toUser:
                i.balance += float(trans.amount)
                to_user_in_new_list = True

        # If not included in the account_list, add.
        if not touser_inlist and not to_user_in_new_list:
            new_added_accounts_list.append(
                Account(trans.toUser, float(trans.amount)))
    logger.info(f"Actions: {actions}")
    for action in actions:
        for account in account_list:
            if action[0] == account.Address:
                alread_in = True if account in edited_accounts else False
                the_account = edited_accounts[edited_accounts.index(account)] if account in edited_accounts else account                
                if the_account.Address == block.fee_address:
                    logger.info(f"Fee Address Input: {the_account.dump_json()}")
                if action[1] == "balance":
                    balance_decimal = Decimal(str(the_account.balance)) + Decimal(str(action[2]))
                    the_account.balance = float(balance_decimal)
                elif action[1] == "sequence_number":
                    the_account.sequence_number += action[2]

                if the_account.Address == block.fee_address:
                    logger.info(f"Fee Address Output: {the_account.dump_json()}")
                if not alread_in:
                    edited_accounts.append(the_account)

    # Syncs new sorted list to block.validating_list

    block.validating_list = sorted(temp_validating_list,
                                   key=lambda x: x.signature)

    new_added_accounts_list = sorted(new_added_accounts_list,
                                     key=lambda x: x.Address)


    logger.debug(f"SaveAccounts list: {new_added_accounts_list + edited_accounts}")
    logger.debug(f"SaveAccounts path: {the_TEMP_ACCOUNTS_PATH}")
    the_account_list_result = SaveAccounts(new_added_accounts_list + edited_accounts,
                 the_TEMP_ACCOUNTS_PATH, sequence=block.sequence_number+block.empty_block_number)

    return [block, the_account_list_result]



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import os
from hashlib import sha256

from naruno.config import PENDING_TRANSACTIONS_PATH
from naruno.lib.config_system import get_config
from naruno.transactions.pending.save_pending import pendingtransactions_db
from naruno.lib.kot import KOT

def DeletePending(tx, custom_PENDING_TRANSACTIONS_PATH=None):
    if custom_PENDING_TRANSACTIONS_PATH == PENDING_TRANSACTIONS_PATH:
        custom_PENDING_TRANSACTIONS_PATH = None    
    file_name = sha256((tx.signature).encode("utf-8")).hexdigest()
    pendingtransactions_db.delete(file_name)  if custom_PENDING_TRANSACTIONS_PATH is None else KOT("pendingtransactions"+custom_PENDING_TRANSACTIONS_PATH, folder=get_config()["main_folder"] + "/db").delete(file_name)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import copy
import json
import time

from naruno.lib.log import clear_logs
from naruno.lib.log import get_logger
from naruno.lib.settings_system import the_settings
from naruno.transactions.transaction import Transaction
import naruno
logger = get_logger("BLOCKCHAIN")


class Block:
    """
    Block class is most important class. It is responsible for
    resetting and saving blocks.

    You must give a creator of the block. This creator will
    own all the coins.
    """

    def __init__(
        self,
        creator,
        previous_hash="1a00d983803e3adcbda2ed40ecba828083221648a90150267d8b0fd500c59750",
    ):
        self.coin_amount = 10000000
        self.first_time = True
        self.creator = creator
        self.genesis_time = int(time.time())
        self.start_time = int(time.time())
        self.block_time = 27

        self.previous_hash = previous_hash
        self.sequence_number = 0
        self.empty_block_number = 0
        self.hard_block_number = 2
        self.gap_block_number = self.hard_block_number + 2

        self.validating_list = []
        self.transaction_fee = 0.02
        self.default_transaction_fee = 0.02
        self.default_optimum_transaction_number = 200
        self.default_increase_of_fee = 0.01
        self.transaction_delay_time = 3600
        self.max_data_size = 1000000

        self.part_amount = 1000

        self.hash = None
        self.part_amount_cache = previous_hash

        self.max_tx_number = 1000
        self.minumum_transfer_amount = 1000

        self.round_1_time = 20
        self.round_1 = False

        self.round_2_starting_time = None
        self.round_2_time = 5
        self.round_2 = False

        self.consensus_timer = 1

        self.validated = False
        self.validated_time = None

        self.dowload_true_block = ""
        self.sync = False

        self.just_one_tx = True

        self.candidate_blocks_check = 80
        self.candidate_blocks_hashes_check = 80
        self.find_validated = 50
        self.process_candidate_blocks_hashes = 80


        self.shares = []
        self.fee_address = creator

        
        if the_settings()["funtionaltest_mode"]:
            self.max_tx_number = 2
            self.max_data_size = 1000

            self.candidate_blocks_check = 50
            self.candidate_blocks_hashes_check = 50
            self.find_validated = 50
            self.process_candidate_blocks_hashes = 50

        

    def reset_the_block(self, custom_nodes=None):
        """
        When the block is verified and if block have a transaction
        and if block have at least half of the max_tx_number transaction,it saves the block
        and makes the edits for the new block.
        """

        self.start_time = (self.genesis_time +
                           ((self.sequence_number + self.empty_block_number) *
                            self.block_time)) + self.block_time

        self.round_1 = False

        self.round_2_starting_time = None
        self.round_2 = False

        self.validated = False
        self.validated_time = None

       
        block2 = copy.copy(self)
        # Resetting and setting the new elements.
        self.previous_hash = self.hash
        self.sequence_number = self.sequence_number + 1
        clear_logs()
        self.validating_list = []
        self.hash = None
        logger.info("New block created")
        self.sync_empty_blocks()
        logger.debug(self.__dict__)
              
        return [block2, self]


    def sync_empty_blocks(self):
        self.start_time = self.genesis_time + ((self.sequence_number + self.empty_block_number) *self.block_time)
               

    def dump_json(self):
        """
        Dumps the block as json.
        """
        temp_block = copy.copy(self)

        temp_validating_list = [
            transaction.dump_json()
            for transaction in temp_block.validating_list
        ]

        temp_block.validating_list = temp_validating_list
        return temp_block.__dict__

    @staticmethod
    def load_json(json_string):
        temp_validating_list = [
            Transaction.load_json(tx) for tx in json_string["validating_list"]
        ]

        the_block_json = json.loads(json.dumps(json_string))
        the_block_json["validating_list"] = temp_validating_list
        the_block = Block("Naruno")
        the_block.__dict__ = the_block_json

        return the_block
